# Copyright 2014-2015 Canonical Limited.
#
# This file is part of charm-helpers.
#
# charm-helpers is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.


import inspect

from docutils.parsers.rst import directives
from sphinx.ext.autosummary import Autosummary
from sphinx.ext.autosummary import get_import_prefixes_from_env
from sphinx.ext.autosummary import import_by_name


class AutoMemberSummary(Autosummary):
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    has_content = True
    option_spec = {
        'toctree': directives.unchanged,
        'nosignatures': directives.flag,
        'template': directives.unchanged,
    }

    def get_items(self, names):
        env = self.state.document.settings.env
        prefixes = get_import_prefixes_from_env(env)

        items = []
        prefix = ''
        shorten = ''

        def _get_items(name):
            _items = super(AutoMemberSummary, self).get_items([shorten + name])
            if self.result.data and ".. deprecated::" in self.result.data[0]:
                # don't show deprecated classes / functions in summary
                return
            for dn, sig, summary, rn in _items:
                if ".. deprecated::" in summary:
                    # don't show deprecated methods in summary
                    continue
                items.append(('%s%s' % (prefix, dn), sig, summary, rn))

        for name in names:
            if '~' in name:
                prefix, name = name.split('~')
                shorten = '~'
            else:
                prefix = ''
                shorten = ''

            try:
                real_name, obj, parent, _ = import_by_name(name, prefixes=prefixes)
            except ImportError:
                self.warn('failed to import %s' % name)
                continue

            if not inspect.ismodule(obj):
                _get_items(name)
                continue

            for member in dir(obj):
                if member.startswith('_'):
                    continue
                mobj = getattr(obj, member)
                if hasattr(mobj, '__module__'):
                    if not mobj.__module__.startswith(real_name):
                        continue  # skip imported classes & functions
                elif hasattr(mobj, '__name__'):
                    continue  # skip imported modules
                else:
                    continue  # skip instances
                _get_items('%s.%s' % (mobj.__module__, member))

        return items


def setup(app):
    app.add_directive('automembersummary', AutoMemberSummary)



# Copyright 2017 Canonical Limited.
#
# This file is part of charms.reactive.
#
# charms.reactive is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charms.reactive is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import json
from collections import UserDict
from itertools import chain

from charmhelpers.core import hookenv, unitdata
from charms.reactive.flags import set_flag, clear_flag, toggle_flag, is_flag_set
from charms.reactive.helpers import data_changed
from charms.reactive.relations import RelationFactory, relation_factory


__all__ = [
    'Endpoint',
]


class Endpoint(RelationFactory):
    """
    New base class for creating interface layers.

    This class is intended to create drop-in, backwards-compatible replacements
    for interface layers previously written using the old
    :class:`~charms.reactive.relations.RelationBase` base class.  With the
    advantages of: having commonly used internal flags managed automatically,
    providing a cleaner, more easily understood pattern for interacting with
    relation data, and being able to use ``@when`` rather than ``@hook`` so
    that interface layers are more similar to charm layers and to remove one
    of the biggest barriers to upgrading from a non-reactive version of a
    charm to a reactive version.

    Four flags are automatically managed for each endpoint. Endpoint handlers
    can react to these flags using the :class:`~charms.reactive.decorators`.

      * ``endpoint.{endpoint_name}.joined`` is set when the endpoint is
        :meth:`joined`: when the first remote unit from any relationship
        connected to this endpoint joins. It is cleared when the last unit
        from all relationships connected to this endpoint departs.
      * ``endpoint.{endpoint_name}.changed`` when any relation data has
        changed. It isn't automatically cleared.
      * ``endpoint.{endpoint_name}.changed.{field}`` when a specific field
        has changed. It isn't automatically cleared.
      * ``endpoint.{endpoint_name}.departed`` when a remote unit is leaving.
         It isn't automatically cleared.

    For the flags that are not automatically cleared, it is up to the interface
    author to clear the flag when it is "handled". The following diagram shows
    how these flags relate. In summary, the ``joined`` flag represents the state
    of the relationship and will be automatically cleared when all units are
    gone. ``changed`` and ``departed`` represents relationship events and have to
    be cleared manually by the handler.

    .. image:: _static/endpoints-workflow.svg

    These flags should only be used by the decorators of the endpoint handlers.
    While it is possible to use them with any decorators in any layer, these
    flags should be considered internal, private implementation details. It is
    the interface layers responsibility to manage and document the public flags
    that make up part of its API.

    Endpoint handlers can iterate over the list of joined relations for an
    endpoint via the :attr:`~charms.reactive.endpoints.Endpoint.relations`
    collection.
    """

    _endpoints = {}

    @classmethod
    def from_name(cls, endpoint_name):
        """
        Return an Endpoint subclass instance based on the name of the endpoint.
        """
        return cls._endpoints.get(endpoint_name)

    @classmethod
    def from_flag(cls, flag):
        """
        Return an Endpoint subclass instance based on the given flag.

        The instance that is returned depends on the endpoint name embedded
        in the flag.  Flags should be of the form ``endpoint.{name}.extra...``,
        though for legacy purposes, the ``endpoint.`` prefix can be omitted.
        The ``{name}}`` portion will be passed to
        :meth:`~charms.reactive.endpoints.Endpoint.from_name`.

        If the flag is not set, an appropriate Endpoint subclass cannot be
        found, or the flag name can't be parsed, ``None`` will be returned.
        """
        if not is_flag_set(flag) or '.' not in flag:
            return None
        parts = flag.split('.')
        if parts[0] == 'endpoint':
            return cls.from_name(parts[1])
        else:
            # some older handlers might not use the 'endpoint' prefix
            return cls.from_name(parts[0])

    @classmethod
    def _startup(cls):
        """
        Create Endpoint instances and manage automatic flags.
        """
        for endpoint_name in sorted(hookenv.relation_types()):
            # populate context based on attached relations
            relf = relation_factory(endpoint_name)
            if not relf or not issubclass(relf, cls):
                continue

            rids = sorted(hookenv.relation_ids(endpoint_name))
            # ensure that relation IDs have the endpoint name prefix, in case
            # juju decides to drop it at some point
            rids = ['{}:{}'.format(endpoint_name, rid) if ':' not in rid
                    else rid for rid in rids]
            endpoint = relf(endpoint_name, rids)
            cls._endpoints[endpoint_name] = endpoint
            endpoint.register_triggers()
            endpoint._manage_departed()
            endpoint._manage_flags()
            for relation in endpoint.relations:
                hookenv.atexit(relation._flush_data)

    def __init__(self, endpoint_name, relation_ids=None):
        self._endpoint_name = endpoint_name
        self._relations = KeyList(map(Relation, relation_ids or []),
                                  key_attr='relation_id')
        self._all_joined_units = None
        self._all_departed_units = None

    def register_triggers(self):
        """
        Called once and only once for each named instance of this endpoint,
        before the endpoint's automatic flags are updated.

        This gives the endpoint implementation a chance to register triggers
        that will honor changes to the automatically managed flags.
        """
        pass

    @property
    def endpoint_name(self):
        """
        Relation name of this endpoint.
        """
        return self._endpoint_name

    @property
    def relations(self):
        """
        Collection of :class:`Relation` instances that are established for
        this :class:`Endpoint`.

        This is a :class:`KeyList`, so it can be iterated and indexed as a list,
        or you can look up relations by their ID.  For example::

            rel0 = endpoint.relations[0]
            assert rel0 is endpoint.relations[rel0.relation_id]
            assert all(rel is endpoint.relations[rel.relation_id]
                       for rel in endpoint.relations)
            print(', '.join(endpoint.relations.keys()))
        """
        return self._relations

    @property
    def is_joined(self):
        """
        Whether this endpoint has remote applications attached to it.
        """
        return len(self.all_units) > 0

    @property
    def joined(self):
        """
        .. deprecated:: 0.6.3
           Use :attr:`is_joined` instead
        """
        return self.is_joined

    def expand_name(self, flag):
        """
        Complete a flag for this endpoint by expanding the endpoint name.

        If the flag does not already contain ``{endpoint_name}``, it will be
        prefixed with ``endpoint.{endpoint_name}.``. Then, any occurance of
        ``{endpoint_name}`` will be replaced with ``self.endpoint_name``.
        """
        if '{endpoint_name}' not in flag:
            flag = 'endpoint.{endpoint_name}.' + flag
        # use replace rather than format to prevent any other braces or braced
        # strings from being touched
        return flag.replace('{endpoint_name}', self.endpoint_name)

    def _manage_departed(self):
        hook_name = hookenv.hook_name()
        rel_hook = hook_name.startswith(self.endpoint_name + '-relation-')
        departed_hook = rel_hook and hook_name.endswith('-departed')
        if not departed_hook:
            return
        relation = self.relations[hookenv.relation_id()]
        unit = RelatedUnit(relation, hookenv.remote_unit())
        self.all_departed_units.append(unit)
        if not relation.joined_units:
            del self.relations[relation.relation_id]

    def _manage_flags(self):
        """
        Manage automatic relation flags.

        Internal use only.
        """
        already_joined = is_flag_set(self.expand_name('joined'))
        hook_name = hookenv.hook_name()
        rel_hook = hook_name.startswith(self.endpoint_name + '-relation-')
        departed_hook = rel_hook and hook_name.endswith('-departed')

        toggle_flag(self.expand_name('joined'), self.is_joined)

        if departed_hook:
            set_flag(self.expand_name('departed'))
        elif self.is_joined:
            clear_flag(self.expand_name('departed'))

        if already_joined and not rel_hook:
            # skip checking relation data outside hooks for this relation
            # to save on API calls to the controller (unless we didn't have
            # the joined flag before, since then we might migrating to Endpoints)
            return

        for unit in self.all_units:
            for key, value in unit.received.items():
                data_key = 'endpoint.{}.{}.{}.{}'.format(self.endpoint_name,
                                                         unit.relation.relation_id,
                                                         unit.unit_name,
                                                         key)
                if data_changed(data_key, value):
                    set_flag(self.expand_name('changed'))
                    set_flag(self.expand_name('changed.{}'.format(key)))
        self.manage_flags()

    def manage_flags(self):
        """
        Method that subclasses can override to perform any flag management
        needed during startup.

        This will be called automatically after the framework-managed automatic
        flags have been updated.
        """
        pass

    @property
    def all_units(self):
        """
        .. deprecated:: 0.6.1
           Use :attr:`all_joined_units` instead
        """
        return self.all_joined_units

    @property
    def all_joined_units(self):
        """
        A list view of all the units of all relations attached to this
        :class:`~charms.reactive.endpoints.Endpoint`.

        This is actually a
        :class:`~charms.reactive.endpoints.CombinedUnitsView`, so the units
        will be in order by relation ID and then unit name, and you can access a
        merged view of all the units' data as a single mapping.  You should be
        very careful when using the merged data collections, however, and
        consider carefully what will happen when the endpoint has multiple
        relations and multiple remote units on each.  It is probably better to
        iterate over each unit and handle its data individually.  See
        :class:`~charms.reactive.endpoints.CombinedUnitsView` for an
        explanation of how the merged data collections work.

        Note that, because a given application might be related multiple times
        on a given endpoint, units may show up in this collection more than
        once.
        """
        if self._all_joined_units is None:
            units = chain.from_iterable(rel.units for rel in self.relations)
            self._all_joined_units = CombinedUnitsView(units)
        return self._all_joined_units

    @property
    def all_departed_units(self):
        """
        Collection of all units that were previously part of any relation on
        this endpoint but which have since departed.

        This collection is persistent and mutable.  The departed units will
        be kept until they are explicitly removed, to allow for reasonable
        cleanup of units that have left.

        Example: You need to run a command each time a unit departs the relation.

        .. code-block:: python

            @when('endpoint.{endpoint_name}.departed')
            def handle_departed_unit(self):
                for name, unit in self.all_departed_units.items():
                    # run the command to remove `unit` from the cluster
                    #  ..
                self.all_departed_units.clear()
                clear_flag(self.expand_name('departed'))

        Once a unit is departed, it will no longer show up in
        :attr:`all_joined_units`.  Note that units are considered departed as
        soon as the departed hook is entered, which differs slightly from how
        the Juju primitives behave (departing units are still returned from
        ``related-units`` until after the departed hook is complete).

        This collection is a :class:`KeyList`, so can be used as a mapping to
        look up units by their unit name, or iterated or accessed by index.
        """
        if self._all_departed_units is None:
            self._all_departed_units = CachedKeyList.load(
                'reactive.endpoints.departed.{}'.format(self.endpoint_name),
                RelatedUnit._deserialize,
                'unit_name')
        return self._all_departed_units


class Relation:
    def __init__(self, relation_id):
        self._relation_id = relation_id
        self._endpoint_name = relation_id.split(':')[0]
        self._application_name = None
        self._units = None
        self._departed_units = None
        self._data = None
        self._app_data = None
        self._remote_app_data = None

    @property
    def relation_id(self):
        """
        This relation's relation ID.
        """
        return self._relation_id

    @property
    def endpoint_name(self):
        """
        This relation's endpoint name.

        This will be the same as the
        :class:`~charms.reactive.endpoints.Endpoint`'s endpoint name.
        """
        return self._endpoint_name

    @property
    def endpoint(self):
        """
        This relation's :class:`Endpoint` instance.
        """
        return Endpoint.from_name(self.endpoint_name)

    @property
    def application_name(self):
        """
        The name of the remote application for this relation, or ``None``.

        This is equivalent to::

            relation.units[0].unit_name.split('/')[0]
        """
        if self._application_name is None and self.units:
            self._application_name = self.units[0].unit_name.split('/')[0]
        return self._application_name

    @property
    def units(self):
        """
        .. deprecated:: 0.6.1
           Use :attr:`joined_units` instead
        """
        return self.joined_units

    @property
    def joined_units(self):
        """
        A list view of all the units joined on this relation.

        This is actually a
        :class:`~charms.reactive.endpoints.CombinedUnitsView`, so the units
        will be in order by unit name, and you can access a merged view of all
        of the units' data with ``self.units.received`` and
        ``self.units.received``.  You should be very careful when using the
        merged data collections, however, and consider carefully what will
        happen when there are multiple remote units.  It is probabaly better to
        iterate over each unit and handle its data individually.  See
        :class:`~charms.reactive.endpoints.CombinedUnitsView` for an
        explanation of how the merged data collections work.

        The view can be iterated and indexed as a list, or you can look up units
        by their unit name.  For example::

            by_index = relation.units[0]
            by_name = relation.units['unit/0']
            assert by_index is by_name
            assert all(unit is relation.units[unit.unit_name]
                       for unit in relation.units)
            print(', '.join(relation.units.keys()))
        """
        if self._units is None:
            self._units = CombinedUnitsView([
                RelatedUnit(self, unit_name) for unit_name in
                sorted(hookenv.related_units(self.relation_id))
            ])
        return self._units

    @property
    def to_publish(self):
        """
        This is the relation data that the local unit publishes so it is
        visible to all related units. Use this to communicate with related
        units. It is a writeable
        :class:`~charms.reactive.endpoints.JSONUnitDataView`.

        All values stored in this collection will be automatically JSON
        encoded when they are published. This means that they need to be JSON
        serializable! Mappings stored in this collection will be encoded with
        sorted keys, to ensure that the encoded representation will only change
        if the actual data changes.

        Changes to this data are published at the end of a succesfull hook. The
        data is reset when a hook fails.
        """
        if self._data is None:
            self._data = JSONUnitDataView(
                hookenv.relation_get(unit=hookenv.local_unit(),
                                     rid=self.relation_id),
                writeable=True)
        return self._data

    @property
    def to_publish_app(self):
        """
        This is the relation data that the local app publishes so it is
        visible to all related units. Use this to communicate with related
        apps. It is a writeable
        :class:`~charms.reactive.endpoints.JSONUnitDataView`.

        Only the leader can set the app-level relation data.

        All values stored in this collection will be automatically JSON
        encoded when they are published. This means that they need to be JSON
        serializable! Mappings stored in this collection will be encoded with
        sorted keys, to ensure that the encoded representation will only change
        if the actual data changes.

        Changes to this data are published at the end of a succesfull hook. The
        data is reset when a hook fails.
        """
        if self._app_data is None:
            # using JSONUnitDataView though it's name includes unit.
            self._app_data = JSONUnitDataView(
                hookenv.relation_get(app=hookenv.application_name(),
                                     rid=self.relation_id),
                writeable=True)
        return self._app_data

    @property
    def to_publish_raw(self):
        """
        This is the raw relation data that the local unit publishes so it is
        visible to all related units. It is a writeable
        :class:`~charms.reactive.endpoints.UnitDataView`. **Only use this
        for backwards compatibility with interfaces that do not use JSON
        encoding.** Use
        :attr:`~charms.reactive.endpoints.Relation.to_publish` instead.

        Changes to this data are published at the end of a succesfull hook. The
        data is reset when a hook fails.
        """
        return self.to_publish.data

    @property
    def to_publish_app_raw(self):
        """
        This is the raw relation data that the app publishes so it is
        visible to all related units. It is a writeable (by the leader only)
        :class:`~charms.reactive.endpoints.UnitDataView`. **Only use this
        for backwards compatibility with interfaces that do not use JSON
        encoding.** Use
        :attr:`~charms.reactive.endpoints.Relation.to_publish` instead.

        Changes to this data are published at the end of a succesfull hook. The
        data is reset when a hook fails.
        """
        return self.to_publish_app.data

    @property
    def received_app(self):
        """
        A :class:`~charms.reactive.endpoints.JSONUnitDataView` of the app-level
        data received from this remote unit over the relation, with values
        being automatically decoded as JSON.
        """
        if self._remote_app_data is None:
            # using JSONUnitDataView though it's name includes unit.
            self._remote_app_data = JSONUnitDataView(hookenv.relation_get(
                app=self.application_name,
                rid=self.relation_id))
        return self._remote_app_data

    @property
    def received_app_raw(self):
        """
        A :class:`~charms.reactive.endpoints.UnitDataView` of the raw app-level
        data received from this remote unit over the relation.
        """
        return self.received_app.raw_data

    def _flush_data(self):
        """
        If this relation's local unit data has been modified, publish it on the
        relation. This should be automatically called.
        """
        if self._data and self._data.modified:
            hookenv.relation_set(self.relation_id, dict(self.to_publish.data))
        if self._app_data and self._app_data.modified:
            hookenv.relation_set(self.relation_id, dict(self.to_publish_app.data), app=True)

    def _serialize(self):
        return self.relation_id

    @classmethod
    def _deserialize(cls, relation_id):
        endpoint = Endpoint.from_name(relation_id.split(':')[0])
        if relation_id in endpoint.relations:
            # prefer joined relations
            return endpoint.relations[relation_id]
        else:
            # handle broken relations
            return Relation(relation_id)


class RelatedUnit:
    """
    Class representing a remote unit on a relation.
    """
    def __init__(self, relation, unit_name, data=None):
        self._relation = relation
        self._unit_name = unit_name
        self._application_name = unit_name.split('/')[0]
        self._data = data

    @property
    def relation(self):
        """
        The relation to which this unit belongs.
        """
        return self._relation

    @property
    def unit_name(self):
        """
        The name of this unit.
        """
        return self._unit_name

    @property
    def application_name(self):
        """
        The name of the application to which this unit belongs.
        """
        return self._application_name

    @property
    def received(self):
        """
        A :class:`~charms.reactive.endpoints.JSONUnitDataView` of the data
        received from this remote unit over the relation, with values being
        automatically decoded as JSON.
        """
        if self._data is None:
            self._data = JSONUnitDataView(hookenv.relation_get(
                unit=self.unit_name,
                rid=self.relation.relation_id))
        return self._data

    @property
    def received_raw(self):
        """
        A :class:`~charms.reactive.endpoints.UnitDataView` of the raw data
        received from this remote unit over the relation.
        """
        return self.received.raw_data

    def _serialize(self):
        return {
            'relation': self.relation._serialize(),
            'unit_name': self.unit_name,
            'data': dict(self.received_raw),
        }

    @classmethod
    def _deserialize(cls, data):
        return cls(Relation._deserialize(data['relation']),
                   data['unit_name'],
                   JSONUnitDataView(data['data']))


class KeyList(list):
    """
    List that also allows accessing items keyed by an attribute on the items.

    Unlike dicts, the keys don't need to be unique.
    """
    def __init__(self, items, key_attr):
        """
        :param list items: List of items
        :param str key_attr: Attribute to use as the key for mapping access.
        """
        super().__init__(items)
        self._key_attr = key_attr

    def _translate_key(self, key):
        if isinstance(key, int):
            return key
        for i, item in enumerate(self):
            if getattr(item, self._key_attr) == key:
                return i
        raise KeyError(key)

    def __getitem__(self, key):
        """
        Access an item in this :class:`~charms.reactive.endpoints.KeyList` by
        either an integer index or a str key.

        If an integer key is given, it will be used as a list index.

        If a str is given, it will be used as a mapping key.  Since keys may not
        be unique, only the first item matching the given key will be returned.
        """
        return super().__getitem__(self._translate_key(key))

    def __delitem__(self, key):
        super().__delitem__(self._translate_key(key))

    def pop(self, key):
        return super().pop(self._translate_key(key))

    def keys(self):
        """
        Return the keys for all items in this
        :class:`~charms.reactive.endpoints.KeyList`.

        Unlike a dict, the keys are not necessarily unique, so this list may
        contain duplicate values.  The keys will be returned in the order of the
        items in the list.
        """
        return [getattr(item, self._key_attr) for item in self]

    def values(self):
        """
        Return just the values of this list.

        This is equivalent to ``list(keylist)``.
        """
        return list(self)

    def items(self):
        return ((getattr(item, self._key_attr), item) for item in self)

    def __contains__(self, key_or_value):
        return key_or_value in self.keys() or key_or_value in self.values()


class CachedKeyList(KeyList):
    """
    Variant of :class:`KeyList` where items are serialized and persisted
    or removed from the persisted copy, whenever the list is modified.
    """
    def __init__(self, cache_key, items, key_attr):
        self._cache_key = cache_key
        super().__init__(items, key_attr)

    @classmethod
    def load(cls, cache_key, deserializer, key_attr):
        """
        Load the persisted cache and return a new instance of this class.
        """
        items = unitdata.kv().get(cache_key) or []
        return cls(cache_key,
                   [deserializer(item) for item in items],
                   key_attr)

    def _save(self):
        if not self:
            unitdata.kv().unset(self._cache_key)
        else:
            unitdata.kv().set(self._cache_key, [item._serialize() for item in self])

    def __setitem__(self, key, value):
        super().__setitem__(self._translate_key(key), value)
        self._save()

    def __delitem__(self, key):
        super().__delitem__(self._translate_key(key))
        self._save()

    def pop(self, key):
        super().pop(self._translate_key(key))
        self._save()

    def remove(self, value):
        super().remove(value)
        self._save()

    def clear(self):
        super().clear()
        self._save()

    def append(self, value):
        super().append(value)
        self._save()

    def extend(self, values):
        super().extend(values)
        self._save()


class CombinedUnitsView(KeyList):
    """
    A :class:`~charms.reactive.endpoints.KeyList` view of
    :class:`~charms.reactive.endpoints.RelatedUnit` items, with properties to
    access a merged view of all of the units' data.

    You can iterate over this view like any other list, or you can look up units
    by their ``unit_name``.  Units will be in order by relation ID and unit
    name.  If a given unit name occurs more than once, accessing it by
    ``unit_name`` will return the one from the lowest relation ID::

        # given the following relations...
        {
            'endpoint:1': {
                'unit/1': {
                    'key0': 'value0_1_1',
                    'key1': 'value1_1_1',
                },
                'unit/0': {
                    'key0': 'value0_1_0',
                    'key1': 'value1_1_0',
                },
            },
            'endpoint:0': {
                'unit/1': {
                    'key0': 'value0_0_1',
                    'key2': 'value2_0_1',
                },
            },
        }

        from_all = endpoint.all_units['unit/1']
        by_rel = endpoint.relations['endpoint:0'].units['unit/1']
        by_index = endpoint.relations[0].units[1]
        assert from_all is by_rel
        assert by_rel is by_index

    You can also use the
    :attr:`~charms.reactive.endpoints.CombinedUnitsView.received` or
    :attr:`~charms.reactive.endpoints.CombinedUnitsView.received_raw`
    properties just like you would on a single unit.  The data in these
    collections will have all of the data from every unit, with units with the
    lowest relation ID and unit name taking precedence if multiple units have
    set a given field.  For example::

        # given the same relations as above...

        # the values across all relations would be:
        assert endpoint.all_units.received['key0'] == 'value0_0_0'
        assert endpoint.all_units.received['key1'] == 'value1_1_0'
        assert endpoint.all_units.received['key2'] == 'value2_0_1'

        # across individual relations:
        assert endpoint.relations[0].units.received['key0'] == 'value0_0_1'
        assert endpoint.relations[0].units.received['key1'] == None
        assert endpoint.relations[0].units.received['key2'] == 'value2_0_1'
        assert endpoint.relations[1].units.received['key0'] == 'value0_1_0'
        assert endpoint.relations[1].units.received['key1'] == 'value1_1_0'
        assert endpoint.relations[1].units.received['key2'] == None

        # and of course you an access them by individual unit
        assert endpoint.relations['endpoint:1'].units['unit/1'].received['key0'] \
                == 'value0_1_1'

    """
    def __init__(self, items):
        super().__init__(sorted(items, key=lambda i: (i.relation.relation_id,
                                                      i.unit_name)),
                         key_attr='unit_name')

    @property
    def received(self):
        """
        Combined :class:`~charms.reactive.endpoints.JSONUnitDataView` of the
        data of all units in this list, with automatic JSON decoding.
        """
        if not hasattr(self, '_data'):
            # NB: units are reversed so that lowest numbered unit takes precedence
            self._data = JSONUnitDataView({key: value
                                           for unit in reversed(self)
                                           for key, value in unit.received_raw.items()})

        return self._data

    @property
    def received_raw(self):
        """
        Combined :class:`~charms.reactive.endpoints.UnitDataView` of the raw data
        of all units in this list, as raw strings.
        """
        return self.received.raw_data


class UnitDataView(UserDict):
    """
    View of a dict containing a unit's data.

    This is like a ``defaultdict(lambda: None)`` which cannot be modified by
    default.
    """
    def __init__(self, data, writeable=False):
        self.data = data
        self._writeable = writeable
        self._modified = False

    @property
    def modified(self):
        """
        Whether this collection has been modified.
        """
        return self._modified

    @property
    def writeable(self):
        """
        Whether this collection can be modified.
        """
        return self._writeable

    def get(self, key, default=None):
        return self.data.get(key, default)

    def __getitem__(self, key):
        return self.data.get(key)

    def __setitem__(self, key, value):
        if not self._writeable:
            raise ValueError('Remote unit data cannot be modified')
        self._modified = True
        self.data[key] = value

    def setdefault(self, key, value):
        if key not in self:
            self[key] = value
        return self[key]


class JSONUnitDataView(UserDict):
    """
    View of a dict that performs automatic JSON en/decoding of items.

    Like :class:`~charms.reactive.endpoints.UnitDataView`, this is like a
    ``defaultdict(lambda: None)`` which cannot be modified by default.

    When decoding, if a value fails to decode, it will just return the raw value
    as a string.

    When encoding, it ensures that keys are sorted to maintain stable and
    consistent encoded representations.

    The original data, without automatic encoding / decoding, can be accessed as
    :attr:`raw_data`.
    """
    def __init__(self, data, writeable=False):
        self.data = UnitDataView(data, writeable)

    @property
    def raw_data(self):
        """
        The data for this collection without automatic encoding / decoding.

        This is an :class:`~charms.reactive.endpoints.UnitDataView` instance.
        """
        return self.data

    @property
    def modified(self):
        """
        Whether this collection has been modified.
        """
        return self.raw_data.modified

    @property
    def writeable(self):
        """
        Whether this collection can be modified.
        """
        return self.raw_data.writeable

    def get(self, key, default=None):
        if key not in self.raw_data:
            return default
        return self[key]

    def __getitem__(self, key):
        value = self.raw_data[key]
        if not value:
            return value
        try:
            return json.loads(value)
        except Exception:
            # Catch json.JSONDecodeError when we drop Python 3.4 support.
            return value

    def __setitem__(self, key, value):
        self.raw_data[key] = json.dumps(value, sort_keys=True)

    def setdefault(self, key, value):
        if key not in self:
            self[key] = value
        return self[key]


hookenv.atstart(Endpoint._startup)



# Copyright 2014-2015 Canonical Limited.
#
# This file is part of charm-helpers.
#
# charm-helpers is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import re
import os
import mock
import shutil
import tempfile
import unittest

from charmhelpers.core import unitdata
from charms import reactive


class TestReactiveHelpers(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_db_dir = tempfile.mkdtemp()
        test_db = os.path.join(cls.test_db_dir, 'test-state.db')
        unitdata._KV = cls.kv = unitdata.Storage(test_db)
        if not hasattr(cls, 'assertItemsEqual'):
            cls.assertItemsEqual = cls.assertCountEqual

    @classmethod
    def tearDownClass(cls):
        cls.kv.close()
        unitdata._KV = None
        shutil.rmtree(cls.test_db_dir)

    def tearDown(self):
        self.kv.cursor.execute('delete from kv')

    def test_toggle_state(self):
        reactive.toggle_state('foo', True)
        reactive.toggle_state('foo', True)
        reactive.toggle_state('bar', False)
        reactive.toggle_state('bar', False)
        assert reactive.is_state('foo')
        assert not reactive.is_state('bar')

    def test_is_state(self):
        assert not reactive.is_state('foo')
        reactive.set_flag('foo')
        assert reactive.is_state('foo')

    def test_all_flags(self):
        reactive.set_flag('foo')
        reactive.set_flag('bar')
        assert reactive.all_flags_set('foo')
        assert reactive.all_flags_set('bar')
        assert reactive.all_flags_set('foo', 'bar')
        assert not reactive.all_flags_set('foo', 'bar', 'qux')
        assert not reactive.all_flags_set('foo', 'qux')
        assert not reactive.all_flags_set('bar', 'qux')
        assert not reactive.all_flags_set('qux')

    def test_any_flags(self):
        reactive.set_flag('foo')
        reactive.set_flag('bar')
        assert reactive.any_flags_set('foo')
        assert reactive.any_flags_set('bar')
        assert reactive.any_flags_set('foo', 'bar')
        assert reactive.any_flags_set('foo', 'bar', 'qux')
        assert reactive.any_flags_set('foo', 'qux')
        assert reactive.any_flags_set('bar', 'qux')
        assert not reactive.any_flags_set('qux')

    def test_expand_replacements(self):
        er = reactive.helpers._expand_replacements
        pat = re.compile(r'{([^}]+)}')
        self.assertItemsEqual(er(pat, lambda v: [v], ['A']), ['A'])
        self.assertItemsEqual(er(pat, lambda v: [v], ['{A}']), ['A'])
        self.assertItemsEqual(er(pat, lambda v: v.split(','), ['{A,B}']), ['A', 'B'])
        self.assertItemsEqual(er(pat, lambda v: v.split(','), ['{A,B}', '{C,D}']), ['A', 'B', 'C', 'D'])
        self.assertItemsEqual(er(pat, lambda v: v.split(','), ['{A,B}{C,D}']), ['AC', 'BC', 'AD', 'BD'])
        self.assertItemsEqual(er(pat, lambda v: v.split(','), ['{A,B}{A,B}']), ['AA', 'BA', 'AB', 'BB'])

    @mock.patch('charmhelpers.core.hookenv.metadata')
    @mock.patch('charmhelpers.core.hookenv.hook_name')
    def test_any_hook(self, hook_name, metadata):
        hook_name.return_value = 'config-changed'
        metadata.return_value = {}
        assert not reactive.helpers.any_hook('foo', 'bar')
        assert reactive.helpers.any_hook('foo', 'config-changed')
        assert reactive.helpers.any_hook('foo', 'config-{set,changed}')
        assert reactive.helpers.any_hook('foo', 'config-{changed,set}')
        assert reactive.helpers.any_hook('foo', '{config,option}-{changed,set}')

        metadata.return_value = {
            'requires': {
                'db1': {'interface': 'mysql'},
                'db2': {'interface': 'postgres'},
            },
            'provides': {
                'db3': {'interface': 'mysql'},
            },
        }
        hook_name.return_value = 'db1-relation-changed'
        assert not reactive.helpers.any_hook('{requires:http}-relation-changed')
        assert not reactive.helpers.any_hook('{requires:postgres}-relation-changed')
        assert reactive.helpers.any_hook('{requires:mysql}-relation-changed')
        hook_name.return_value = 'db3-relation-changed'
        assert not reactive.helpers.any_hook('{requires:mysql}-relation-changed')
        assert reactive.helpers.any_hook('{provides:mysql}-relation-changed')
        assert reactive.helpers.any_hook('{provides:mysql}-relation-{joined,changed}')

    @mock.patch('charmhelpers.core.host.file_hash')
    def test_any_file_changed(self, file_hash):
        self.kv.update({
            'file1': 'hash1',
            'file2': 'hash2',
            'file3': 'hash3',
        }, prefix='reactive.files_changed.')
        file_hash.side_effect = [
            # file2 and file3 change
            'hash1',
            'hash1', 'changed1',
            'changed2',

            # no changes
            'hash1',
            'hash1', 'changed1',
            'changed2',
        ]
        afc = reactive.helpers.any_file_changed
        assert not afc(['file1'])
        assert afc(['file1', 'file2'])
        assert afc(['file3'], hash_type='sha256')

        assert not afc(['file1'])
        assert not afc(['file1', 'file2'])
        assert not afc(['file3'])

        self.assertEqual(file_hash.call_args_list, [
            mock.call('file1', hash_type='md5'),
            mock.call('file1', hash_type='md5'),
            mock.call('file2', hash_type='md5'),
            mock.call('file3', hash_type='sha256'),
            mock.call('file1', hash_type='md5'),
            mock.call('file1', hash_type='md5'),
            mock.call('file2', hash_type='md5'),
            mock.call('file3', hash_type='md5'),
        ])

    @mock.patch('charmhelpers.core.host.file_hash')
    def test_any_file_changed_argtypes(self, file_hash):
        file_hash.return_value = 'beep'
        # A filename may be a callable, in which case it is called and
        # the result used, and are cast to strings.
        reactive.helpers.any_file_changed(['one', lambda: 'two', 3])
        file_hash.assert_has_calls([mock.call('one', hash_type='md5'),
                                    mock.call('two', hash_type='md5'),
                                    mock.call('3', hash_type='md5')])

    def test_was_invoked(self):
        assert not reactive.helpers.was_invoked('foo')
        assert not reactive.helpers.was_invoked('foo')
        reactive.helpers.mark_invoked('foo')
        assert reactive.helpers.was_invoked('foo')

    def test_data_changed(self):
        assert reactive.helpers.data_changed('foo', {'foo': 'FOO', 'bar': u'\ua000BAR'})
        assert not reactive.helpers.data_changed('foo', {'foo': 'FOO', 'bar': u'\ua000BAR'})
        assert reactive.helpers.data_changed('bar', {'foo': 'FOO', 'bar': u'\ua000BAR'})
        assert reactive.helpers.data_changed('foo', {'foo': 'QUX', 'bar': u'\ua000BAR'})
        assert not reactive.helpers.data_changed('foo', {'foo': 'QUX', 'bar': u'\ua000BAR'})

    @mock.patch.object(reactive.helpers, 'any_hook')
    def test__hook(self, any_hook):
        pats = ['pat1', 'pat2']

        self.kv.set('reactive.dispatch.phase', 'hooks')
        any_hook.return_value = True
        assert reactive.helpers._hook(pats)
        any_hook.assert_called_once_with(*pats)

        any_hook.reset_mock()
        self.kv.set('reactive.dispatch.phase', 'other')
        any_hook.return_value = True
        assert not reactive.helpers._hook(pats)
        assert not any_hook.called

        any_hook.reset_mock()
        self.kv.set('reactive.dispatch.phase', 'hooks')
        any_hook.return_value = False
        assert not reactive.helpers._hook(pats)
        any_hook.assert_called_once_with('pat1', 'pat2')

    def test__when_all(self):
        test = lambda: reactive.helpers._when_all(['state1', 'state2'])

        self.kv.set('reactive.dispatch.phase', 'hooks')
        assert not test(), 'when_all: hooks; none'

        self.kv.set('reactive.dispatch.phase', 'other')
        assert not test(), 'when_all: other; none'

        self.kv.set('reactive.dispatch.phase', 'hooks')
        reactive.set_flag('state1')
        assert not test(), 'when_all: hooks; one'

        self.kv.set('reactive.dispatch.phase', 'other')
        assert not test(), 'when_all: other; one'

        self.kv.set('reactive.dispatch.phase', 'hooks')
        reactive.set_flag('state2')
        assert not test(), 'when_all: hooks; both'

        self.kv.set('reactive.dispatch.phase', 'other')
        assert test(), 'when_all: other; both'

    def test__when_any(self):
        test = lambda: reactive.helpers._when_any(['state1', 'state2'])

        self.kv.set('reactive.dispatch.phase', 'hooks')
        assert not test(), 'when_any: hooks; none'

        self.kv.set('reactive.dispatch.phase', 'other')
        assert not test(), 'when_any: other; none'

        self.kv.set('reactive.dispatch.phase', 'hooks')
        reactive.set_flag('state1')
        assert not test(), 'when_any: hooks; one'

        self.kv.set('reactive.dispatch.phase', 'other')
        assert test(), 'when_any: other; one'

        self.kv.set('reactive.dispatch.phase', 'hooks')
        reactive.set_flag('state2')
        assert not test(), 'when_any: hooks; both'

        self.kv.set('reactive.dispatch.phase', 'other')
        assert test(), 'when_any: other; both'

    def test__when_none(self):
        test = lambda: reactive.helpers._when_none(['state1', 'state2'])

        self.kv.set('reactive.dispatch.phase', 'hooks')
        assert not test(), 'when_not: hooks; none'

        self.kv.set('reactive.dispatch.phase', 'other')
        assert test(), 'when_none: other; none'

        self.kv.set('reactive.dispatch.phase', 'hooks')
        reactive.set_flag('state1')
        assert not test(), 'when_none: hooks; one'

        self.kv.set('reactive.dispatch.phase', 'other')
        assert not test(), 'when_none: other; one'

        self.kv.set('reactive.dispatch.phase', 'hooks')
        reactive.set_flag('state2')
        assert not test(), 'when_none: hooks; both'

        self.kv.set('reactive.dispatch.phase', 'other')
        assert not test(), 'when_none: other; both'

    def test__when_not_all(self):
        test = lambda: reactive.helpers._when_not_all(['state1', 'state2'])

        self.kv.set('reactive.dispatch.phase', 'hooks')
        assert not test(), 'when_not_all: hooks; none'

        self.kv.set('reactive.dispatch.phase', 'other')
        assert test(), 'when_not_all: other; none'

        self.kv.set('reactive.dispatch.phase', 'hooks')
        reactive.set_flag('state1')
        assert not test(), 'when_not_all: hooks; one'

        self.kv.set('reactive.dispatch.phase', 'other')
        assert test(), 'when_not_all: other; one'

        self.kv.set('reactive.dispatch.phase', 'hooks')
        reactive.set_flag('state2')
        assert not test(), 'when_not_all: hooks; both'

        self.kv.set('reactive.dispatch.phase', 'other')
        assert not test(), 'when_not_all: other; both'

    @mock.patch('charmhelpers.core.hookenv.hook_name')
    def test_restricted_hook(self, hook_name):
        self.kv.set('reactive.dispatch.phase', 'restricted')
        hook_name.return_value = 'meter-status-changed'
        assert not reactive.helpers._restricted_hook('bar')
        assert not reactive.helpers._restricted_hook('config-changed')
        assert reactive.helpers._restricted_hook('meter-status-changed')

        self.kv.set('reactive.dispatch.phase', 'other')
        hook_name.return_value = 'meter-status-changed'
        assert not reactive.helpers._restricted_hook('bar')
        assert not reactive.helpers._restricted_hook('config-changed')
        assert not reactive.helpers._restricted_hook('meter-status-changed')



# Copyright 2019 Canonical Limited.
#
# This file is part of charms.reactive.
#
# charms.reactive is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import json
from mock import Mock, patch

from charms.reactive import (
    Field,
    BaseRequest,
    BaseResponse,
    RequesterEndpoint,
    ResponderEndpoint,
)


class TResponse(BaseResponse):
    actual_foo = Field('The fooness actually provided')
    actual_bar = Field('The barness actually provided')


class TRequest(BaseRequest):
    RESPONSE_CLASS = TResponse
    foo = Field('The fooness requested')
    bar = Field('The barness requested')


class TRequester(RequesterEndpoint):
    REQUEST_CLASS = TRequest

    rel = Mock(name='req_rel',
               spec=['to_publish', 'joined_units'],
               to_publish={})
    unit = Mock(name='req_unit',
                spec=['relation', 'received'],
                relation=rel,
                received=rel.to_publish)

    _endpoint_name = 'requester'
    _relations = [rel]
    _all_joined_units = []

    def __init__(self):
        # patch out Endpoint's init so that we just use our mock relation data
        # but can still test the RequesterEndpoint init logic
        with patch('charms.reactive.endpoints.Endpoint.__init__'):
            super().__init__()


class TResponder(ResponderEndpoint):
    REQUEST_CLASS = TRequest

    rel = Mock(name='res_rel',
               spec=['to_publish', 'joined_units'],
               to_publish={})
    unit = Mock(name='res_unit',
                spec=['relation', 'received'],
                relation=rel,
                received=rel.to_publish)

    _endpoint_name = 'responder'
    _relations = [rel]
    _all_joined_units = []

    def __init__(self, requester):
        # patch out Endpoint's init so that we just use our mock relation data
        # but can still test the ResponderEndpoint init logic
        with patch('charms.reactive.endpoints.Endpoint.__init__'):
            super().__init__()


def test_request_response():
    # hook up the relations so that each side sees the other side's unit
    TRequester.rel.joined_units = [TResponder.unit]
    TRequester._all_joined_units = [TResponder.unit]
    TResponder.rel.joined_units = [TRequester.unit]
    TResponder._all_joined_units = [TRequester.unit]

    requester = TRequester()

    assert len(requester.requests) == 0
    assert len(requester.responses) == 0

    req1 = TRequest.create(requester.rel, foo='foo', bar='bar')
    req2 = TRequest.create(requester.rel, foo='unfoo', bar='unbar')
    assert len(requester.requests) == 2
    assert len(requester.responses) == 0
    assert len(req1.request_id) > 0
    assert len(req2.request_id) > 0
    assert TRequest.find(foo='foo') is req1
    assert TRequest.find(bar='unbar') is req2
    assert TRequest.find(foo='other') is None

    responder = TResponder(requester)
    assert len(responder.all_requests) == 2
    assert len(responder.new_requests) == 2

    rreq1 = TRequest.find(foo='foo')
    rreq2 = TRequest.find(foo='unfoo')
    rreq1.create_response(actual_foo='FOO', actual_bar='BAR')
    rreq2.create_response(actual_foo='unfoo', actual_bar='unbar')
    assert len(responder.all_requests) == 2
    assert len(responder.new_requests) == 0

    assert len(requester.requests) == 2
    assert len(requester.responses) == 2

    res1 = requester.response_by_foo('foo')
    res2 = requester.response_by_bar('unbar')
    assert res1
    assert res2
    assert requester.response_by_foo('other') is None
    assert requester.response_by_foo('FOO') is None
    assert (res1.actual_foo, res1.actual_bar) == ('FOO', 'BAR')
    assert (res2.actual_foo, res2.actual_bar) == ('unfoo', 'unbar')

    # verify that they can be serialized
    assert json.dumps(req1) != '{}'
    assert json.dumps(res1) != '{}'



# Copyright 2014-2017 Canonical Limited.
#
# This file is part of charms.reactive.
#
# charms.reactive is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import os
import mock
import shutil
import tempfile
import unittest

from charmhelpers.core import hookenv
from charmhelpers.core import unitdata
from charms import reactive


class TestReactiveDecorators(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_db_dir = tempfile.mkdtemp()
        test_db = os.path.join(cls.test_db_dir, 'test-state.db')
        unitdata._KV = cls.kv = unitdata.Storage(test_db)
        cls._log = mock.patch('charmhelpers.core.hookenv.log')
        cls.log = cls._log.start()
        cls._charm_dir = mock.patch('charmhelpers.core.hookenv.charm_dir', lambda: 'charm_dir')
        cls._charm_dir.start()
        if not hasattr(cls, 'assertItemsEqual'):
            cls.assertItemsEqual = cls.assertCountEqual

    @classmethod
    def tearDownClass(cls):
        cls._log.stop()
        cls._charm_dir.stop()
        cls.kv.close()
        unitdata._KV = None
        shutil.rmtree(cls.test_db_dir)

    def tearDown(self):
        reactive.bus.Handler.clear()
        self.kv.cursor.execute('delete from kv')

    @mock.patch.object(hookenv, 'relation_type')
    @mock.patch.object(reactive.decorators, 'endpoint_from_name')
    @mock.patch.object(reactive.decorators, '_hook')
    def test_hook(self, _hook, from_name, relation_type):
        _hook.return_value = True
        from_name.return_value = 'RB.from_name'
        relation_type.return_value = 'rel_type'
        action = mock.Mock(name='action')

        @reactive.hook('{requires:mysql}-relation-{joined,changed}')
        def test_action(*args):
            action(*args)

        handler = reactive.bus.Handler.get(test_action)
        assert handler.test()
        handler.invoke()

        _hook.assert_called_once_with(('{requires:mysql}-relation-{joined,changed}',))
        from_name.assert_called_once_with('rel_type')
        action.assert_called_once_with('RB.from_name')

        delattr(handler, '_args_evaled')
        action.reset_mock()
        from_name.return_value = None
        handler.invoke()
        action.assert_called_once_with()

    @mock.patch.object(reactive.decorators, 'endpoint_from_flag')
    @mock.patch.object(reactive.decorators, '_action_id')
    @mock.patch.object(reactive.decorators, '_when_all')
    def test_when_all(self, _when_all, _action_id, from_flag):
        reactive.bus.Handler._CONSUMED_FLAGS.clear()
        _when_all.return_value = True
        _action_id.return_value = 'f:l:test_action'
        from_flag.side_effect = [None, 'rel', None]
        action = mock.Mock(name='action')

        @reactive.when_all('foo', 'bar', 'qux')
        def test_action(*args):
            action(*args)

        handler = reactive.bus.Handler.get(test_action)
        assert handler.test()
        handler.invoke()

        _when_all.assert_called_once_with(('foo', 'bar', 'qux'))
        self.assertEqual(from_flag.call_args_list, [
            mock.call('foo'),
            mock.call('bar'),
            mock.call('qux'),
        ])
        action.assert_called_once_with('rel')
        self.assertEqual(reactive.bus.Handler._CONSUMED_FLAGS, set(['foo', 'bar', 'qux']))

        action.reset_mock()
        assert handler.test()
        handler.invoke()
        action.assert_called_once_with('rel')

    @mock.patch.object(reactive.decorators, 'when_all')
    def test_when(self, when_all):
        @reactive.when('foo', 'bar', 'qux')
        def test_action(*args):
            pass
        when_all.assert_called_once_with('foo', 'bar', 'qux')

    @mock.patch.object(reactive.decorators, 'endpoint_from_flag')
    @mock.patch.object(reactive.decorators, '_action_id')
    @mock.patch.object(reactive.decorators, '_when_any')
    def test_when_any(self, _when_any, _action_id, from_flag):
        reactive.bus.Handler._CONSUMED_FLAGS.clear()
        _when_any.return_value = True
        _action_id.return_value = 'f:l:test_action'
        from_flag.side_effect = [None, 'rel', None]
        action = mock.Mock(name='action')

        @reactive.when_any('foo', 'bar', 'qux')
        def test_action(*args):
            action(*args)

        handler = reactive.bus.Handler.get(test_action)
        assert handler.test()
        handler.invoke()

        _when_any.assert_called_once_with(('foo', 'bar', 'qux'))
        assert not from_flag.called
        action.assert_called_once_with()
        self.assertEqual(reactive.bus.Handler._CONSUMED_FLAGS, set(['foo', 'bar', 'qux']))

    @mock.patch.object(reactive.decorators, 'endpoint_from_flag')
    @mock.patch.object(reactive.decorators, '_action_id')
    @mock.patch.object(reactive.decorators, '_when_none')
    def test_when_none(self, _when_none, _action_id, from_flag):
        reactive.bus.Handler._CONSUMED_FLAGS.clear()
        _when_none.return_value = True
        _action_id.return_value = 'f:l:test_action'
        from_flag.return_value = 'rel'
        action = mock.Mock(name='action')

        @reactive.when_none('foo', 'bar', 'qux')
        def test_action():
            action()

        handler = reactive.bus.Handler.get(test_action)
        assert handler.test()
        handler.invoke()

        _when_none.assert_called_once_with(('foo', 'bar', 'qux'))
        assert not from_flag.called
        action.assert_called_once_with()
        self.assertEqual(reactive.bus.Handler._CONSUMED_FLAGS, set(['foo', 'bar', 'qux']))

    @mock.patch.object(reactive.decorators, 'when_none')
    def test_when_not(self, when_none):
        @reactive.when_not('foo', 'bar', 'qux')
        def test_action(*args):
            pass
        when_none.assert_called_once_with('foo', 'bar', 'qux')

    @mock.patch.object(reactive.decorators, 'endpoint_from_flag')
    @mock.patch.object(reactive.decorators, '_action_id')
    @mock.patch.object(reactive.decorators, '_when_not_all')
    def test_when_not_all(self, _when_not_all, _action_id, from_flag):
        reactive.bus.Handler._CONSUMED_FLAGS.clear()
        _when_not_all.return_value = True
        _action_id.return_value = 'f:l:test_action'
        from_flag.return_value = 'rel'
        action = mock.Mock(name='action')

        @reactive.when_not_all('foo', 'bar', 'qux')
        def test_action():
            action()

        handler = reactive.bus.Handler.get(test_action)
        assert handler.test()
        handler.invoke()

        _when_not_all.assert_called_once_with(('foo', 'bar', 'qux'))
        assert not from_flag.called
        action.assert_called_once_with()
        self.assertEqual(reactive.bus.Handler._CONSUMED_FLAGS, set(['foo', 'bar', 'qux']))

    @mock.patch.object(reactive.decorators, 'any_file_changed')
    def test_when_file_changed(self, any_file_changed):
        any_file_changed.side_effect = [True, False]

        @reactive.decorators.when_file_changed('file1', 'file2')
        def test_action1():
            pass

        @reactive.decorators.when_file_changed('file3', 'file4', hash_type='sha256')
        def test_action2():
            pass

        handler1 = reactive.bus.Handler.get(test_action1)
        handler2 = reactive.bus.Handler.get(test_action2)
        assert handler1.test()
        assert not handler2.test()
        self.assertEqual(any_file_changed.call_args_list, [
            mock.call(('file1', 'file2')),
            mock.call(('file3', 'file4'), hash_type='sha256'),
        ])

    @mock.patch('charmhelpers.core.hookenv.log')
    def test_not_unless(self, log):
        action = mock.Mock()

        def log_msg(num):
            return log.call_args_list[num][0][0]

        @reactive.not_unless('foo', 'bar')
        def test():
            """Doc string."""
            action()

        self.assertEqual(test.__doc__, 'Doc string.')

        test()
        reactive.set_flag('foo')
        test()
        reactive.set_flag('bar')
        test()

        self.assertEqual(action.call_count, 3)
        assert log_msg(0).endswith('test called before flags: foo, bar'), log_msg(0)
        assert log_msg(1).endswith('test called before flag: bar'), log_msg(1)
        self.assertIn('tests/test_decorators.py:', reactive.bus._action_id(test))
        self.assertIn(':test', reactive.bus._action_id(test))
        self.assertIn('tests/test_decorators.py:', reactive.bus._short_action_id(test))
        self.assertIn(':test', reactive.bus._short_action_id(test))

    def test_only_once(self):
        calls = []

        @reactive.decorators.only_once
        def test():
            calls.append(len(calls) + 1)

        handler = reactive.bus.Handler.get(test)

        assert handler.test()
        handler.invoke()
        assert not handler.test()
        self.assertEquals(calls, [1])

    def test_only_once_parens(self):
        calls = []

        @reactive.decorators.only_once()
        def test():
            calls.append(len(calls) + 1)

        handler = reactive.bus.Handler.get(test)

        assert handler.test()
        handler.invoke()
        assert not handler.test()
        self.assertEquals(calls, [1])

    def test_multi(self):
        action1 = mock.Mock(name='action1')
        action2 = mock.Mock(name='action2')
        action3 = mock.Mock(name='action3')

        @reactive.when('foo')
        @reactive.when('bar')
        def test1():
            action1()

        @reactive.when('foo')
        @reactive.when_not('bar')
        def test2():
            action2()

        reactive.set_flag('foo')
        reactive.set_flag('bar')
        reactive.bus.dispatch()
        assert action1.called
        assert not action2.called

        action1.reset_mock()
        reactive.remove_state('bar')
        reactive.bus.dispatch()
        assert not action1.called
        assert action2.called

        @reactive.when('foo')
        def test3():
            action3()
            reactive.remove_state('bar')

        reactive.set_flag('bar')
        action2.reset_mock()
        reactive.bus.dispatch()
        assert action3.called
        assert action2.called  # should be called on second iteration

    @mock.patch.object(reactive.decorators, '_restricted_hook')
    def test_collect_metrics(self, _restricted_hook):
        _restricted_hook.return_value = True
        action = mock.Mock(name='action')

        @reactive.collect_metrics()
        def test_action(*args):
            action(*args)

        handler = reactive.bus.Handler.get(test_action)
        assert handler.test()
        handler.invoke()

        _restricted_hook.assert_called_once_with('collect-metrics')
        action.assert_called_once()

    @mock.patch.object(reactive.decorators, '_restricted_hook')
    def test_meter_status_changed(self, _restricted_hook):
        _restricted_hook.return_value = True
        action = mock.Mock(name='action')

        @reactive.meter_status_changed()
        def test_action(*args):
            action(*args)

        handler = reactive.bus.Handler.get(test_action)
        assert handler.test()
        handler.invoke()

        _restricted_hook.assert_called_once_with('meter-status-changed')
        action.assert_called_once()



# Copyright 2014-2017 Canonical Limited.
#
# This file is part of charms.reactive.
#
# charms.reactive is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import sys
import mock
import tempfile
import unittest
from pathlib import Path

from charmhelpers.core import unitdata
from charms.reactive import (
    Endpoint,
    set_flag,
    is_flag_set,
    clear_flag,
    register_trigger,
)
from charms.reactive.bus import discover, dispatch, Handler


class TestEndpoint(unittest.TestCase):
    def setUp(self):
        tests_dir = Path(__file__).parent

        tf = tempfile.NamedTemporaryFile(delete=False)
        tf.close()
        self.test_db = Path(tf.name)
        unitdata._KV = self.kv = unitdata.Storage(str(self.test_db))

        self.log_p = mock.patch('charmhelpers.core.hookenv.log')
        self.log_p.start()

        self.charm_dir = str(tests_dir / 'data')
        self.charm_dir_p = mock.patch('charmhelpers.core.hookenv.charm_dir')
        mcharm_dir = self.charm_dir_p.start()
        mcharm_dir.side_effect = lambda: self.charm_dir

        self.hook_name = 'upgrade-charm'
        self.hook_name_p = mock.patch('charmhelpers.core.hookenv.hook_name')
        mhook_name = self.hook_name_p.start()
        mhook_name.side_effect = lambda: self.hook_name

        self.local_unit_p = mock.patch('charmhelpers.core.hookenv.local_unit',
                                       mock.MagicMock(return_value='local/0'))
        self.local_unit_p.start()

        self.app_name_p = mock.patch('charmhelpers.core.hookenv.application_name',
                                     mock.MagicMock(return_value='local'))
        self.app_name_p.start()

        self.remote_unit = None
        self.remote_unit_p = mock.patch('charmhelpers.core.hookenv.remote_unit')
        mremote_unit = self.remote_unit_p.start()
        mremote_unit.side_effect = lambda: self.remote_unit

        self.relation_id = None
        self.relation_id_p = mock.patch('charmhelpers.core.hookenv.relation_id')
        mrelation_id = self.relation_id_p.start()
        mrelation_id.side_effect = lambda: self.relation_id

        self.relations = {
            'test-endpoint': [
                {
                    'local': {'app-key': 'value'},
                    'local/0': {'key': 'value'},
                    'unit': {'simple': 'value', 'complex': '[1, 2]'},
                    'unit/0': {'foo': 'yes'},
                    'unit/1': {},
                },
                {
                    'local': {},
                    'local/0': {},
                    'unit': {},
                    'unit/0': {'bar': '[1, 2]'},
                    'unit/1': {'foo': 'no'},
                },
            ],
        }

        def _rel(rid):
            rn, ri = rid.split(':')
            return self.relations[rn][int(ri)]

        def _rel_get(attribute=None, unit=None, rid=None, app=None):
            data = _rel(rid)[unit or app]
            if attribute is not None:
                return data[attribute]
            else:
                return data

        self.rel_ids_p = mock.patch('charmhelpers.core.hookenv.relation_ids')
        rel_ids_m = self.rel_ids_p.start()
        rel_ids_m.side_effect = lambda endpoint: [
            '{}:{}'.format(endpoint, i) for i in range(
                len(self.relations.get(endpoint, [])))]
        self.rel_units_p = mock.patch('charmhelpers.core.hookenv.related_units')
        rel_units_m = self.rel_units_p.start()
        rel_units_m.side_effect = lambda rid: [
            key for key in _rel(rid).keys()
            if (not key.startswith('local') and
                '/' in key and  # exclude apps
                not _rel(rid)[key].get('departed'))]
        self.rel_get_p = mock.patch('charmhelpers.core.hookenv.relation_get')
        rel_get_m = self.rel_get_p.start()
        rel_get_m.side_effect = _rel_get

        self.rel_set_p = mock.patch('charmhelpers.core.hookenv.relation_set')
        self.relation_set = self.rel_set_p.start()

        self.data_changed_p = mock.patch('charms.reactive.endpoints.data_changed')
        self.data_changed = self.data_changed_p.start()

        self.atexit_p = mock.patch('charmhelpers.core.hookenv.atexit')
        self.atexit = self.atexit_p.start()

        self.sysm_p = mock.patch.dict(sys.modules)
        self.sysm_p.start()

        self.kv.set('reactive.endpoints.departed.test-endpoint', [
            {
                'relation': 'test-endpoint:1',
                'unit_name': 'unit/3',
                'data': {'departed': 'true'},
            },
            {
                'relation': 'test-endpoint:2',
                'unit_name': 'unit/4',
                'data': {'departed': 'true'},
            },
        ])

        discover()

    def tearDown(self):
        self.log_p.stop()
        self.charm_dir_p.stop()
        self.hook_name_p.stop()
        self.local_unit_p.stop()
        self.remote_unit_p.stop()
        self.rel_ids_p.stop()
        self.rel_units_p.stop()
        self.rel_get_p.stop()
        self.rel_set_p.stop()
        self.data_changed_p.stop()
        self.atexit_p.stop()
        self.test_db.unlink()
        self.sysm_p.stop()
        Endpoint._endpoints.clear()
        Handler._HANDLERS.clear()

    def test_from_name(self):
        Endpoint._endpoints['foo'] = endpoint = Endpoint('foo')

        self.assertIs(Endpoint.from_name('foo'), endpoint)
        self.assertIsNone(Endpoint.from_name('bar'))

    def test_from_flag(self):
        Endpoint._endpoints['foo'] = endpoint = Endpoint('foo')

        self.assertIsNone(Endpoint.from_flag('foo'))
        self.assertIsNone(Endpoint.from_flag('bar.qux.zod'))

        # should return None for unset flag
        self.assertIsNone(Endpoint.from_flag('endpoint.foo.qux'))

        # once flag is set, should return the endpoint
        set_flag('endpoint.foo.qux')
        self.assertIs(Endpoint.from_flag('endpoint.foo.qux'), endpoint)

        set_flag('foo.qux')
        self.assertIs(Endpoint.from_flag('foo.qux'), endpoint)

    def test_startup(self):
        assert not is_flag_set('endpoint.test-endpoint.joined')
        assert not is_flag_set('endpoint.test-endpoint.changed')
        assert not is_flag_set('endpoint.test-endpoint.changed.foo')

        set_flag('endpoint.test-endpoint2.joined')
        set_flag('alias.test-endpoint2.joined')

        def _register_triggers(self):
            joined_flag = self.expand_name('endpoint.{endpoint_name}.joined')
            alias_joined_flag = self.expand_name('alias.{endpoint_name}.joined')
            register_trigger(when=joined_flag, set_flag=alias_joined_flag)
            register_trigger(when_not=joined_flag, clear_flag=alias_joined_flag)

        self.data_changed.return_value = True
        with mock.patch.object(Endpoint, 'register_triggers',
                               _register_triggers):
            Endpoint._startup()
        assert Endpoint.from_name('test-endpoint') is not None
        assert Endpoint.from_name('test-endpoint').endpoint_name == 'test-endpoint'
        assert Endpoint.from_name('test-endpoint').is_joined
        assert Endpoint.from_name('test-endpoint').joined  # deprecated
        assert is_flag_set('endpoint.test-endpoint.joined')
        assert is_flag_set('endpoint.test-endpoint.changed')
        assert is_flag_set('endpoint.test-endpoint.changed.foo')
        assert Endpoint.from_name('test-endpoint2') is not None
        assert Endpoint.from_name('test-endpoint2').endpoint_name == 'test-endpoint2'
        assert not Endpoint.from_name('test-endpoint2').is_joined
        assert not Endpoint.from_name('test-endpoint2').joined  # deprecated
        assert not is_flag_set('endpoint.test-endpoint2.joined')
        assert not is_flag_set('endpoint.test-endpoint2.changed')
        assert not is_flag_set('endpoint.test-endpoint2.changed.foo')
        assert is_flag_set('alias.test-endpoint.joined')
        assert not is_flag_set('alias.test-endpoint2.joined')
        assert not is_flag_set('alias.test-endpoint3.joined')
        self.assertEqual(self.atexit.call_args_list, [
            mock.call(Endpoint.from_name('test-endpoint').relations[0]._flush_data),
            mock.call(Endpoint.from_name('test-endpoint').relations[1]._flush_data),
        ])

        # already joined, not relation hook
        clear_flag('endpoint.test-endpoint.changed')
        clear_flag('endpoint.test-endpoint.changed.foo')
        Endpoint._startup()
        assert not is_flag_set('endpoint.test-endpoint.changed')
        assert not is_flag_set('endpoint.test-endpoint.changed.foo')

        # relation hook
        self.hook_name = 'test-endpoint-relation-joined'
        clear_flag('endpoint.test-endpoint.changed')
        clear_flag('endpoint.test-endpoint.changed.foo')
        Endpoint._startup()
        assert is_flag_set('endpoint.test-endpoint.changed')
        assert is_flag_set('endpoint.test-endpoint.changed.foo')

        # not already joined
        self.hook_name = 'upgrade-charm'
        clear_flag('endpoint.test-endpoint.joined')
        clear_flag('endpoint.test-endpoint.changed')
        clear_flag('endpoint.test-endpoint.changed.foo')
        Endpoint._startup()
        assert is_flag_set('endpoint.test-endpoint.changed')
        assert is_flag_set('endpoint.test-endpoint.changed.foo')

        # data not changed
        self.data_changed.return_value = False
        clear_flag('endpoint.test-endpoint.joined')
        clear_flag('endpoint.test-endpoint.changed')
        clear_flag('endpoint.test-endpoint.changed.foo')
        Endpoint._startup()
        assert not is_flag_set('endpoint.test-endpoint.changed')
        assert not is_flag_set('endpoint.test-endpoint.changed.foo')

    def test_collections(self):
        Endpoint._startup()
        tep = Endpoint.from_name('test-endpoint')

        self.assertEqual(len(tep.relations), 2)
        self.assertEqual(len(tep.relations[0].joined_units), 2)
        self.assertEqual(len(tep.relations[1].joined_units), 2)
        self.assertEqual(tep.relations['test-endpoint:0'].relation_id, 'test-endpoint:0')
        self.assertEqual(len(tep.all_joined_units), 4)
        self.assertEqual([u.unit_name for r in tep.relations for u in r.joined_units],
                         ['unit/0', 'unit/1', 'unit/0', 'unit/1'])
        self.assertEqual([u.unit_name for u in tep.all_joined_units],
                         ['unit/0', 'unit/1', 'unit/0', 'unit/1'])
        self.assertEqual(tep.relations[0].joined_units['unit/1'].unit_name, 'unit/1')
        self.assertEqual(tep.relations[0].relation_id, 'test-endpoint:0')
        self.assertEqual(tep.relations[0].endpoint_name, 'test-endpoint')
        self.assertEqual(tep.relations[0].application_name, 'unit')
        self.assertEqual(tep.relations[0].joined_units[0].unit_name, 'unit/0')
        self.assertEqual(tep.all_joined_units.keys(), ['unit/0', 'unit/1', 'unit/0', 'unit/1'])
        self.assertEqual(tep.relations[0].joined_units.keys(), ['unit/0', 'unit/1'])
        self.assertEqual(tep.relations.keys(), ['test-endpoint:0', 'test-endpoint:1'])
        self.assertIs(tep.all_units, tep.all_joined_units)  # deprecated
        self.assertIs(tep.relations[0].units, tep.relations[0].joined_units)  # deprecated

    def test_departed(self):
        # clean up some units for this test
        del self.relations['test-endpoint'][0]['unit/1']
        del self.relations['test-endpoint'][1]['unit/0']
        # add unit/2 as joined
        self.relations['test-endpoint'][0]['unit/2'] = {'departed': 'yes'}
        # but set current hook as unit/2 departing
        self.hook_name = 'test-endpoint-relation-departed'
        self.relation_id = 'test-endpoint:0'
        self.remote_unit = 'unit/2'
        Endpoint._startup()
        tep = Endpoint.from_name('test-endpoint')

        self.assertCountEqual(tep.relations.keys(), ['test-endpoint:0', 'test-endpoint:1'])
        self.assertCountEqual(tep.all_joined_units.keys(), ['unit/0', 'unit/1'])
        self.assertCountEqual(tep.all_departed_units.keys(), ['unit/2', 'unit/3', 'unit/4'])
        self.assertEqual(tep.all_departed_units['unit/2'].received_raw['departed'], 'yes')
        self.assertIs(tep.all_departed_units['unit/3'].received['departed'], True)

        self.assertCountEqual(self.kv.get('reactive.endpoints.departed.test-endpoint'), [
            {
                'relation': 'test-endpoint:0',
                'unit_name': 'unit/2',
                'data': {
                    'departed': 'yes',
                },
            },
            {
                'relation': 'test-endpoint:1',
                'unit_name': 'unit/3',
                'data': {
                    'departed': 'true',
                },
            },
            {
                'relation': 'test-endpoint:2',
                'unit_name': 'unit/4',
                'data': {
                    'departed': 'true',
                },
            },
        ])
        del tep.all_departed_units['unit/3']
        self.assertCountEqual(self.kv.get('reactive.endpoints.departed.test-endpoint'), [
            {
                'relation': 'test-endpoint:0',
                'unit_name': 'unit/2',
                'data': {
                    'departed': 'yes',
                },
            },
            {
                'relation': 'test-endpoint:2',
                'unit_name': 'unit/4',
                'data': {
                    'departed': 'true',
                },
            },
        ])
        del tep.all_departed_units['unit/2']
        del tep.all_departed_units['unit/4']
        self.assertIsNone(self.kv.get('reactive.endpoints.departed.test-endpoint'))

        # test relation moves to broken during last departed hook
        self.relation_id = 'test-endpoint:1'
        self.remote_unit = 'unit/1'
        self.relations['test-endpoint'][1]['unit/1'] = {'departed': 'yes'}
        Endpoint._startup()
        tep = Endpoint.from_name('test-endpoint')
        self.assertEqual(tep.relations.keys(), ['test-endpoint:0'])

    def test_receive(self):
        Endpoint._startup()
        tep = Endpoint.from_name('test-endpoint')

        self.assertEqual(tep.all_joined_units.received_raw, {'foo': 'yes',
                                                             'bar': '[1, 2]'})
        self.assertEqual(tep.all_joined_units.received, {'foo': 'yes',
                                                         'bar': [1, 2]})
        self.assertEqual(tep.relations[0].joined_units.received_raw, {'foo': 'yes'})
        self.assertEqual(tep.relations[1].joined_units.received_raw, {'foo': 'no',
                                                                      'bar': '[1, 2]'})
        self.assertEqual(tep.relations[0].joined_units.received, {'foo': 'yes'})
        self.assertEqual(tep.relations[1].joined_units.received, {'foo': 'no',
                                                                  'bar': [1, 2]})
        self.assertEqual(tep.relations[0].joined_units[0].received_raw, {'foo': 'yes'})
        self.assertEqual(tep.relations[0].joined_units[1].received_raw, {})
        self.assertEqual(tep.relations[1].joined_units[0].received_raw, {'bar': '[1, 2]'})
        self.assertEqual(tep.relations[1].joined_units[1].received_raw, {'foo': 'no'})
        self.assertEqual(tep.relations[0].joined_units[0].received, {'foo': 'yes'})
        self.assertEqual(tep.relations[0].joined_units[1].received, {})
        self.assertEqual(tep.relations[1].joined_units[0].received, {'bar': [1, 2]})
        self.assertEqual(tep.relations[1].joined_units[1].received, {'foo': 'no'})

        self.assertEqual(tep.all_joined_units.received_raw['bar'], '[1, 2]')
        self.assertEqual(tep.all_joined_units.received_raw.get('bar'), '[1, 2]')
        self.assertEqual(tep.all_joined_units.received['bar'], [1, 2])
        self.assertEqual(tep.all_joined_units.received.get('bar'), [1, 2])
        self.assertIsNone(tep.all_joined_units.received_raw['none'])
        self.assertEqual(tep.all_joined_units.received_raw.get('none', 'default'), 'default')
        self.assertIsNone(tep.all_joined_units.received['none'])
        self.assertEqual(tep.all_joined_units.received.get('none', 'default'), 'default')

        assert not tep.all_joined_units.received_raw.writeable
        assert not tep.all_joined_units.received.writeable

        with self.assertRaises(ValueError):
            tep.all_joined_units.received_raw['foo'] = 'nope'

        with self.assertRaises(ValueError):
            tep.relations[0].joined_units.received_raw['foo'] = 'nope'

        with self.assertRaises(ValueError):
            tep.relations[0].joined_units[0].received_raw['foo'] = 'nope'

        with self.assertRaises(ValueError):
            tep.all_joined_units.received['foo'] = 'nope'

        with self.assertRaises(ValueError):
            tep.relations[0].joined_units.received['foo'] = 'nope'

        with self.assertRaises(ValueError):
            tep.relations[0].joined_units[0].received['foo'] = 'nope'

    def test_receive_app(self):
        Endpoint._startup()
        tep = Endpoint.from_name('test-endpoint')

        self.assertEqual(tep.relations[0].received_app_raw, {'simple': 'value', 'complex': '[1, 2]'})
        self.assertEqual(tep.relations[1].received_app_raw, {})
        self.assertEqual(tep.relations[0].received_app, {'simple': 'value', 'complex': [1, 2]})
        self.assertEqual(tep.relations[1].received_app, {})

        assert not tep.relations[0].received_app_raw.writeable
        assert not tep.relations[1].received_app_raw.writeable
        assert not tep.relations[0].received_app.writeable
        assert not tep.relations[1].received_app.writeable

        with self.assertRaises(ValueError):
            tep.relations[0].received_app_raw['simple'] = 'nope'

        with self.assertRaises(ValueError):
            tep.relations[1].received_app_raw['simple'] = 'nope'

        with self.assertRaises(ValueError):
            tep.relations[0].received_app['simple'] = 'nope'

        with self.assertRaises(ValueError):
            tep.relations[1].received_app['simple'] = 'nope'

    def test_to_publish(self):
        Endpoint._startup()
        tep = Endpoint.from_name('test-endpoint')
        rel = tep.relations[0]

        self.assertEqual(rel.to_publish_raw, {'key': 'value'})
        rel._flush_data()
        assert not self.relation_set.called

        rel.to_publish_raw['key'] = 'new-value'
        rel._flush_data()
        self.relation_set.assert_called_once_with('test-endpoint:0', {'key': 'new-value'})

        self.relation_set.reset_mock()
        rel.to_publish['key'] = {'new': 'complex'}
        rel._flush_data()
        self.relation_set.assert_called_once_with('test-endpoint:0', {'key': '{"new": "complex"}'})

        rel.to_publish_raw.update({'key': 'new-new'})
        self.assertEqual(rel.to_publish_raw, {'key': 'new-new'})

        rel.to_publish.update({'key': {'new': 'new'}})
        self.assertEqual(rel.to_publish_raw, {'key': '{"new": "new"}'})

        assert 'foo' not in rel.to_publish
        assert rel.to_publish.get('foo', 'one') == 'one'
        assert 'foo' not in rel.to_publish
        assert rel.to_publish.setdefault('foo', 'two') == 'two'
        assert 'foo' in rel.to_publish
        assert rel.to_publish['foo'] == 'two'
        del rel.to_publish['foo']
        assert 'foo' not in rel.to_publish
        with self.assertRaises(KeyError):
            del rel.to_publish['foo']
        assert 'foo' not in rel.to_publish
        assert rel.to_publish['foo'] is None

    def test_to_publish_app(self):
        Endpoint._startup()
        tep = Endpoint.from_name('test-endpoint')
        rel = tep.relations[0]

        self.assertEqual(rel.to_publish_app_raw, {'app-key': 'value'})
        rel._flush_data()
        assert not self.relation_set.called

        rel.to_publish_app_raw['app-key'] = 'new-value'
        rel._flush_data()
        self.relation_set.assert_called_once_with('test-endpoint:0', {'app-key': 'new-value'}, app=True)

        self.relation_set.reset_mock()
        rel.to_publish_app['app-key'] = {'new': 'complex'}
        rel._flush_data()
        self.relation_set.assert_called_once_with('test-endpoint:0', {'app-key': '{"new": "complex"}'}, app=True)

        rel.to_publish_app.update({'app-key': 'new-new'})
        self.assertEqual(rel.to_publish_app, {'app-key': 'new-new'})

        rel.to_publish_app.update({'app-key': {'new': 'new'}})
        self.assertEqual(rel.to_publish_app, {'app-key': {"new": "new"}})

        assert 'foo' not in rel.to_publish
        assert rel.to_publish.get('foo', 'one') == 'one'
        assert 'foo' not in rel.to_publish
        assert rel.to_publish.setdefault('foo', 'two') == 'two'
        assert 'foo' in rel.to_publish
        assert rel.to_publish['foo'] == 'two'
        del rel.to_publish['foo']
        assert 'foo' not in rel.to_publish
        with self.assertRaises(KeyError):
            del rel.to_publish['foo']
        assert 'foo' not in rel.to_publish
        assert rel.to_publish['foo'] is None

    def test_handlers(self):
        Handler._HANDLERS = {k: h for k, h in Handler._HANDLERS.items()
                             if hasattr(h, '_action') and
                             h._action.__qualname__.startswith('TestAltRequires.')}
        assert Handler._HANDLERS
        preds = [h._predicates[0].args[0][0] for h in Handler.get_handlers()]
        for pred in preds:
            self.assertRegex(pred, r'^endpoint.test-endpoint.')

        self.data_changed.return_value = False
        Endpoint._startup()
        tep = Endpoint.from_name('test-endpoint')

        self.assertCountEqual(tep.invocations, [])
        dispatch()
        self.assertCountEqual(tep.invocations, [
            'joined: test-endpoint',
        ])

        tep.invocations.clear()
        clear_flag('endpoint.test-endpoint.joined')
        clear_flag('endpoint.test-endpoint.changed')
        clear_flag('endpoint.test-endpoint.changed.foo')
        clear_flag('endpoint.test-endpoint2.joined')
        clear_flag('endpoint.test-endpoint2.changed')
        clear_flag('endpoint.test-endpoint2.changed.foo')
        self.data_changed.return_value = True
        Endpoint._startup()
        dispatch()
        self.assertCountEqual(tep.invocations, [
            'joined: test-endpoint',
            'changed: test-endpoint',
            'changed.foo: test-endpoint',
        ])

        tep.invocations.clear()
        clear_flag('endpoint.test-endpoint.joined')
        clear_flag('endpoint.test-endpoint.changed')
        clear_flag('endpoint.test-endpoint.changed.foo')
        clear_flag('endpoint.test-endpoint2.joined')
        clear_flag('endpoint.test-endpoint2.changed')
        clear_flag('endpoint.test-endpoint2.changed.foo')
        self.relations['test-endpoint2'] = [
            {
                'unit/0': {'foo': 'yes'},
                'unit/1': {},
            },
            {
                'unit/0': {},
                'unit/1': {'foo': 'no'},
            },
        ]
        Endpoint._startup()
        dispatch()
        self.assertCountEqual(tep.invocations, [
            'joined: test-endpoint',
            'joined: test-endpoint2',
            'changed: test-endpoint',
            'changed: test-endpoint2',
            'changed.foo: test-endpoint',
            'changed.foo: test-endpoint2',
        ])



# Copyright 2014-2015 Canonical Limited.
#
# This file is part of charm-helpers.
#
# charm-helpers is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import os
import shlex

from charmhelpers.cli import cmdline
from charmhelpers.core import templating
from charms.reactive import helpers
from charms.reactive import bus


@cmdline.subcommand()
@cmdline.test_command
def hook(*hook_patterns):
    """
    Check if the current hook matches one of the patterns.
    """
    return helpers._hook(hook_patterns)


@cmdline.subcommand()
@cmdline.test_command
def when(*desired_flags):
    """
    Alias of when_all.
    """
    return helpers._when_all(desired_flags)


@cmdline.subcommand()
@cmdline.test_command
def when_all(*desired_flags):
    """
    Check if all of the desired_flags are active and have changed.
    """
    return helpers._when_all(desired_flags)


@cmdline.subcommand()
@cmdline.test_command
def when_any(*desired_flags):
    """
    Check if any of the desired_flags are active and have changed.
    """
    return helpers._when_any(desired_flags)


@cmdline.subcommand()
@cmdline.test_command
def when_not(*desired_flags):
    """
    Alias of when_none.
    """
    return helpers._when_none(desired_flags)


@cmdline.subcommand()
@cmdline.test_command
def when_none(*desired_flags):
    """
    Check if none of the desired_flags are active and have changed.
    """
    return helpers._when_none(desired_flags)


@cmdline.subcommand()
@cmdline.test_command
def when_not_all(*desired_flags):
    """
    Check if at least one of the desired_flags is not active.
    """
    return helpers._when_not_all(desired_flags)


@cmdline.subcommand()
@cmdline.test_command
def when_file_changed(*filenames):
    """
    Check if files have changed since the last time they were checked.
    """
    return helpers.any_file_changed(filenames)


@cmdline.subcommand()
@cmdline.test_command
def only_once(handler_id):
    """
    Check if handler has already been run in the past.
    """
    return not helpers.was_invoked(handler_id)


@cmdline.subcommand()
@cmdline.no_output
def mark_invoked(*handler_ids):
    """
    Record that the handler has been invoked, for use with only_once.
    """
    for handler_id in handler_ids:
        helpers.mark_invoked(handler_id)


@cmdline.subcommand()  # noqa: C901
def test(*handlers):
    """
    Combined test function to apply one or more tests to multiple handlers.

    Each handler spec given should be a single argument but can contain shell
    quotes to group the parts, and should follow the form:

        'HANDLER_NAME HANDLER_ID [TEST_NAME TEST_ARGS]...'

    Each TEST_ARGS value can have further shell quoting.  For example:

        charms.reactive test 'foo foo_id when "foo.connected foo.available" when_not foo.disabled'
    """
    passed = []
    for handler_spec in handlers:
        handler_name, handler_id, tests = _parse_handler_spec(handler_spec)
        result = True
        states = set()
        for test_name, test_args in tests:
            if test_name == 'hook':
                result &= hook(*test_args)
            elif test_name == 'when':
                result &= when(*test_args)
                states.update(test_args)
            elif test_name == 'when_all':
                result &= when_all(*test_args)
                states.update(test_args)
            elif test_name == 'when_any':
                result &= when_any(*test_args)
                states.update(test_args)
            elif test_name == 'when_not':
                result &= when_not(*test_args)
                states.update(test_args)
            elif test_name == 'when_none':
                result &= when_none(*test_args)
                states.update(test_args)
            elif test_name == 'when_not_all':
                result &= when_not_all(*test_args)
                states.update(test_args)
            elif test_name == 'when_file_changed':
                result &= when_file_changed(*test_args)
            elif test_name == 'only_once':
                result &= only_once(handler_id)
            else:
                raise ValueError('Invalid test: %s' % test_name)
        if states:
            result &= bus.FlagWatch.watch(handler_id, states)
        if result:
            passed.append(handler_name)
    return ','.join(passed)


def _parse_handler_spec(handler_spec):
    parts = shlex.split(handler_spec)
    handler_name, handler_id = parts[:2]
    # one or more pairs of test_name + test_args
    # test_args can be further shell quoted
    tests = zip(parts[2::2], map(shlex.split, parts[3::2]))
    return handler_name, handler_id, tests


@cmdline.subcommand()
@cmdline.no_output
def render_template(source, target):
    """
    Render a Jinja2 template from $CHARM_DIR/templates using the current
    environment variables as the template context.
    """
    templating.render(source, target, os.environ)



import unittest

import mock

from charms.reactive import flags


class TestTriggers(unittest.TestCase):
    def setUp(self):
        kv_p = mock.patch('charmhelpers.core.unitdata.kv')
        kv = kv_p.start()
        self.addCleanup(kv_p.stop)
        kv.return_value = MockKV()

    def test_no_triggers(self):
        assert not flags.any_flags_set('foo', 'bar', 'qux')
        flags.set_flag('foo')
        assert not flags.any_flags_set('bar', 'qux')
        flags.clear_flag('foo')
        assert not flags.any_flags_set('foo', 'bar', 'qux')

    def test_when_set(self):
        flags.register_trigger(when='foo', set_flag='bar')
        flags.set_flag('foo')
        assert flags.is_flag_set('bar')
        flags.clear_flag('foo')
        assert flags.is_flag_set('bar')

    def test_when_clear(self):
        flags.register_trigger(when='foo', clear_flag='qux')
        flags.set_flag('qux')
        flags.set_flag('foo')
        assert not flags.is_flag_set('qux')

    def test_when_not_set_clear(self):
        flags.register_trigger(when_not='foo',
                               set_flag='bar',
                               clear_flag='qux')
        flags.set_flag('noop')
        flags.clear_flag('noop')
        assert not flags.is_flag_set('bar')

        flags.set_flag('foo')
        flags.set_flag('qux')
        assert not flags.is_flag_set('bar')
        flags.clear_flag('foo')
        assert flags.is_flag_set('bar')
        assert not flags.is_flag_set('qux')

    def test_callbacks(self):
        a = mock.Mock()
        b = mock.Mock()
        c = mock.Mock()
        flags.register_trigger(when='a', callback=a)
        flags.register_trigger(when_not='b', callback=b)
        flags.set_flag('a')
        assert a.call_count == 1
        assert b.call_count == 0
        assert c.call_count == 0
        flags.clear_flag('a')
        assert a.call_count == 1
        assert b.call_count == 0
        assert c.call_count == 0
        flags.set_flag('b')
        assert a.call_count == 1
        assert b.call_count == 0
        assert c.call_count == 0
        flags.clear_flag('b')
        assert a.call_count == 1
        assert b.call_count == 1
        assert c.call_count == 0
        flags.set_flag('c')
        assert a.call_count == 1
        assert b.call_count == 1
        assert c.call_count == 0
        flags.clear_flag('c')
        assert a.call_count == 1
        assert b.call_count == 1
        assert c.call_count == 0

    def test_get_unset_flags(self):
        self.assertEqual(flags.get_flags(), [])
        flags.set_flag('foo')
        self.assertEqual(flags.get_flags(), ['foo'])
        self.assertEqual(flags.get_unset_flags('foo', 'bar'), ['bar'])


class MockKV:
    def __init__(self):
        self.data = {}

    def get(self, key, default=None):
        return self.data.get(key, default)

    def getrange(self, prefix, strip=False):
        results = {}
        for key, value in self.data.items():
            if key.startswith(prefix):
                if strip:
                    key = key[len(prefix):]
                results[key] = value
        return results

    def set(self, key, value):
        self.data[key] = value

    def update(self, data, prefix=''):
        for key, value in data.items():
            self.set(prefix + key, value)

    def unset(self, key):
        self.data.pop(key, None)


if __name__ == '__main__':
    unittest.main()



import re
import argparse
import sphinx.ext.autodoc
from charmhelpers.cli import describe_arguments


class CLIDoc(sphinx.ext.autodoc.FunctionDocumenter):
    """
    Automatically generate documentation for CLI entry-points.
    """

    def generate(self, more_content=None, real_modname=None,
                 check_module=False, all_members=False):
        if not self.parse_name():
            # need a module to import
            self.directive.warn(
                'don\'t know which module to import for autodocumenting '
                '%r (try placing a "module" or "currentmodule" directive '
                'in the document, or giving an explicit module name)'
                % self.name)
            return

        # now, import the module and get object to document
        if not self.import_object():
            return

        if not (type(self.object).__name__ == 'function' and
                self.object.__module__ == 'charms.reactive.cli'):
            return super(CLIDoc, self).generate(more_content, real_modname, check_module, all_members)

        parser = argparse.ArgumentParser()
        parser.prog = 'charms.reactive {}'.format(self.object_name)
        for args, kwargs in describe_arguments(self.object):
            parser.add_argument(*args, **kwargs)

        usage = parser.format_usage()
        usage = re.sub('usage: (\S+) (\S+) (.*)', r'\1 **\2** `\3`', usage)
        self.add_line(usage, '<clidoc>')
        self.add_line('', '<clidoc>')
        self.indent += '    '
        self.add_content(more_content)


def setup(app):
    app.add_autodocumenter(CLIDoc)



# Copyright 2014-2017 Canonical Limited.
#
# This file is part of charms.reactive
#
# charms.reactive is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charms.reactive is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

from functools import wraps, partial
from inspect import getfile, isclass, signature
from pathlib import Path

from charmhelpers.core import hookenv
from charms.reactive.bus import Handler
from charms.reactive.bus import _action_id
from charms.reactive.bus import _short_action_id
from charms.reactive.flags import get_flags
from charms.reactive.relations import endpoint_from_name
from charms.reactive.relations import endpoint_from_flag
from charms.reactive.endpoints import Endpoint
from charms.reactive.helpers import _hook
from charms.reactive.helpers import _restricted_hook
from charms.reactive.helpers import _when_all
from charms.reactive.helpers import _when_any
from charms.reactive.helpers import _when_none
from charms.reactive.helpers import _when_not_all
from charms.reactive.helpers import any_file_changed
from charms.reactive.helpers import was_invoked
from charms.reactive.helpers import mark_invoked


__all__ = [
    'when',
    'when_all',
    'when_any',
    'when_not',
    'when_none',
    'when_not_all',
    'not_unless',
    'when_file_changed',
    'collect_metrics',
    'meter_status_changed',
    'only_once',  # DEPRECATED
    'hook',  # DEPRECATED
]


def hook(*hook_patterns):
    """
    Register the decorated function to run when the current hook matches any of
    the ``hook_patterns``.

    This decorator is generally deprecated and should only be used when
    absolutely necessary.

    The hook patterns can use the ``{interface:...}`` and ``{A,B,...}`` syntax
    supported by :func:`~charms.reactive.bus.any_hook`.

    Note that hook decorators **cannot** be combined with :func:`when` or
    :func:`when_not` decorators.
    """
    def _register(action):
        def arg_gen():
            # use a generator to defer calling of hookenv.relation_type, for tests
            rel = endpoint_from_name(hookenv.relation_type())
            if rel:
                yield rel

        handler = Handler.get(action)
        handler.add_predicate(partial(_hook, hook_patterns))
        handler.add_args(arg_gen())
        return action
    return _register


def _when_decorator(predicate, desired_flags, action, legacy_args=False):
    endpoint_names = _get_endpoint_names(action)
    has_relname_flag = _has_endpoint_name_flag(desired_flags)
    params = signature(action).parameters
    has_params = len(params) > 0
    if has_relname_flag and not endpoint_names:
        # If this is an Endpoint handler but there are no endpoints
        # for this interface & role, then we shouldn't register its
        # handlers.  It probably means we only use a different role.
        return action
    for endpoint_name in endpoint_names or [None]:
        handler = Handler.get(action, endpoint_name)
        flags = _expand_endpoint_name(endpoint_name, desired_flags)
        handler.add_predicate(partial(predicate, flags))
        if _is_endpoint_method(action):
            # Endpoint handler methods expect self to be passed in to conform
            # to instance method convention. But mutliple decorators should
            # take care to not pass in multiple copies of self.
            if not handler.has_args:
                handler.add_args(map(endpoint_from_name, [endpoint_name]))
        elif has_params and legacy_args:
            # Handlers should all move to not taking any params and getting
            # the Endpoint instances from the context, but during the
            # transition, we need to provide for handlers expecting args.
            handler.add_args(filter(None, map(endpoint_from_flag, flags)))
        handler.register_flags(flags)
    return action


def when(*desired_flags):
    """
    Alias for :func:`~charms.reactive.decorators.when_all`.
    """
    return when_all(*desired_flags)


def when_all(*desired_flags):
    """
    Register the decorated function to run when all of ``desired_flags`` are active.

    Note that handlers whose conditions match are triggered at least once per
    hook invocation.

    For backwards compatibility, this decorator can pass arguments, but it is
    recommended to use argument-less handlers.  See
    `the summary <#charms-reactive-decorators>`_ for more information.
    """
    return partial(_when_decorator, _when_all, desired_flags, legacy_args=True)


def when_any(*desired_flags):
    """
    Register the decorated function to run when any of ``desired_flags`` are active.

    This decorator will never cause arguments to be passed to the handler.

    Note that handlers whose conditions match are triggered at least once per
    hook invocation.
    """
    return partial(_when_decorator, _when_any, desired_flags, legacy_args=False)


def when_not(*desired_flags):
    """
    Alias for :func:`~charms.reactive.decorators.when_none`.
    """
    return when_none(*desired_flags)


def when_none(*desired_flags):
    """
    Register the decorated function to run when none of ``desired_flags`` are
    active.

    This decorator will never cause arguments to be passed to the handler.

    Note that handlers whose conditions match are triggered at least once per
    hook invocation.
    """
    return partial(_when_decorator, _when_none, desired_flags, legacy_args=False)


def when_not_all(*desired_flags):
    """
    Register the decorated function to run when one or more of the
    ``desired_flags`` are not active.

    This decorator will never cause arguments to be passed to the handler.

    Note that handlers whose conditions match are triggered at least once per
    hook invocation.
    """
    return partial(_when_decorator, _when_not_all, desired_flags, legacy_args=False)


def when_file_changed(*filenames, **kwargs):
    """
    Register the decorated function to run when one or more files have changed.

    :param list filenames: The names of one or more files to check for changes
        (a callable returning the name is also accepted).
    :param str hash_type: The type of hash to use for determining if a file has
        changed.  Defaults to 'md5'.  Must be given as a kwarg.
    """
    def _register(action):
        handler = Handler.get(action)
        handler.add_predicate(partial(any_file_changed, filenames, **kwargs))
        return action
    return _register


def not_unless(*desired_flags):
    """
    Assert that the decorated function can only be called if the desired_flags
    are active.

    Note that, unlike :func:`when`, this does **not** trigger the decorated
    function if the flags match.  It **only** raises an exception if the
    function is called when the flags do not match.

    This is primarily for informational purposes and as a guard clause.
    """
    def _decorator(func):
        action_id = _action_id(func)
        short_action_id = _short_action_id(func)

        @wraps(func)
        def _wrapped(*args, **kwargs):
            active_flags = get_flags()
            missing_flags = [flag for flag in desired_flags if flag not in active_flags]
            if missing_flags:
                hookenv.log('%s called before flag%s: %s' % (
                    short_action_id,
                    's' if len(missing_flags) > 1 else '',
                    ', '.join(missing_flags)), hookenv.WARNING)
            return func(*args, **kwargs)
        _wrapped._action_id = action_id
        _wrapped._short_action_id = short_action_id
        return _wrapped
    return _decorator


def only_once(action=None):
    """
    .. deprecated:: 0.5.0
       Use :func:`when_not` in combination with :func:`set_state` instead. This
       handler is deprecated because it might actually be
       `called multiple times <https://github.com/juju-solutions/charms.reactive/issues/22>`_.

    Register the decorated function to be run once, and only once.

    This decorator will never cause arguments to be passed to the handler.
    """
    if action is None:
        # allow to be used as @only_once or @only_once()
        return only_once

    action_id = _action_id(action)
    handler = Handler.get(action)
    handler.add_predicate(lambda: not was_invoked(action_id))
    handler.add_post_callback(partial(mark_invoked, action_id))
    return action


def collect_metrics():
    """
    Register the decorated function to run for the collect_metrics hook.
    """
    def _register(action):
        handler = Handler.get(action)
        handler.add_predicate(partial(_restricted_hook, 'collect-metrics'))
        return action
    return _register


def meter_status_changed():
    """
    Register the decorated function to run when a meter status change has been detected.
    """
    def _register(action):
        handler = Handler.get(action)
        handler.add_predicate(partial(_restricted_hook, 'meter-status-changed'))
        return action
    return _register


def _has_endpoint_name_flag(flags):
    """
    Detect if the given flags contain any that use ``{endpoint_name}``.
    """
    return '{endpoint_name}' in ''.join(flags)


def _get_endpoint_names(handler):
    filepath = Path(getfile(handler))
    role = filepath.stem
    interface = filepath.parent.name
    if role not in ('requires', 'provides', 'peers'):
        return []
    endpoint_names = hookenv.role_and_interface_to_relations(role, interface)
    if not endpoint_names:
        return []
    return endpoint_names


def _expand_endpoint_name(endpoint_name, flags):
    """
    Populate any ``{endpoint_name}`` tags in the flag names for the given
    handler, based on the handlers module / file name.
    """
    return tuple(flag.format(endpoint_name=endpoint_name) for flag in flags)


def _is_endpoint_method(handler):
    """
    from the context.  Unfortunately, we can't directly detect whether
    a handler is an Endpoint method, because at the time of decoration,
    the class doesn't actually exist yet so it's impossible to get a
    reference to it.  So, we use the heuristic of seeing if the handler
    takes only a single ``self`` param and there is an Endpoint class in
    the handler's globals.
    """
    params = signature(handler).parameters
    has_self = len(params) == 1 and list(params.keys())[0] == 'self'
    has_endpoint_class = any(isclass(g) and issubclass(g, Endpoint)
                             for g in handler.__globals__.values())
    return has_self and has_endpoint_class



# Copyright 2014-2015 Canonical Limited.
#
# This file is part of charm-helpers.
#
# charm-helpers is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import re
import json
import hashlib

from charmhelpers.core import host
from charmhelpers.core import hookenv
from charmhelpers.core import unitdata
from charmhelpers.cli import cmdline
from charms.reactive.flags import any_flags_set, all_flags_set
# import deprecated functions for backwards compatibility
from charms.reactive.flags import is_state, all_states, any_states  # noqa


__all__ = [
    'data_changed',
    'is_data_changed',
    'any_file_changed',
]


def _expand_replacements(pat, subf, values):
    while any(pat.search(r) for r in values):
        new_values = []
        for value in values:
            m = pat.search(value)
            if not m:
                new_values.append(value)
                continue
            whole_match = m.group(0)
            selected_groups = m.groups()
            for replacement in subf(*selected_groups):
                # have to replace one match at a time, or we'll lose combinations
                # e.g., '{A,B}{A,B}' would turn to ['AA', 'BB'] instead of
                # ['A{A,B}', 'B{A,B}'] -> ['AA', 'AB', 'BA', 'BB']
                new_values.append(value.replace(whole_match, replacement, 1))
        values = new_values
    return values


@cmdline.subcommand()
@cmdline.test_command
def any_hook(*hook_patterns):
    """
    Assert that the currently executing hook matches one of the given patterns.

    Each pattern will match one or more hooks, and can use the following
    special syntax:

      * ``db-relation-{joined,changed}`` can be used to match multiple hooks
        (in this case, ``db-relation-joined`` and ``db-relation-changed``).
      * ``{provides:mysql}-relation-joined`` can be used to match a relation
        hook by the role and interface instead of the relation name.  The role
        must be one of ``provides``, ``requires``, or ``peer``.
      * The previous two can be combined, of course: ``{provides:mysql}-relation-{joined,changed}``
    """
    current_hook = hookenv.hook_name()

    # expand {role:interface} patterns
    i_pat = re.compile(r'{([^:}]+):([^}]+)}')
    hook_patterns = _expand_replacements(i_pat, hookenv.role_and_interface_to_relations, hook_patterns)

    # expand {A,B,C,...} patterns
    c_pat = re.compile(r'{((?:[^:,}]+,?)+)}')
    hook_patterns = _expand_replacements(c_pat, lambda v: v.split(','), hook_patterns)

    return current_hook in hook_patterns


def any_file_changed(filenames, hash_type='md5'):
    """
    Check if any of the given files have changed since the last time this
    was called.

    :param list filenames: Names of files to check. Accepts callables returning
        the filename.
    :param str hash_type: Algorithm to use to check the files.
    """
    changed = False
    for filename in filenames:
        if callable(filename):
            filename = str(filename())
        else:
            filename = str(filename)
        old_hash = unitdata.kv().get('reactive.files_changed.%s' % filename)
        new_hash = host.file_hash(filename, hash_type=hash_type)
        if old_hash != new_hash:
            unitdata.kv().set('reactive.files_changed.%s' % filename, new_hash)
            changed = True  # mark as changed, but keep updating hashes
    return changed


def was_invoked(invocation_id):
    """
    Returns whether the given ID has been invoked before, as per :func:`mark_invoked`.

    This is useful for ensuring that a given block only runs one time::

        def foo():
            if was_invoked('foo'):
                return
            do_something()
            mark_invoked('foo')

    This is also available as a decorator at
    :func:`~charms.reactive.decorators.only_once`.
    """
    return unitdata.kv().get('reactive.invoked.%s' % invocation_id, False)


def mark_invoked(invocation_id):
    """
    Mark the given ID as having been invoked, for use with :func:`was_invoked`.
    """
    unitdata.kv().set('reactive.invoked.%s' % invocation_id, True)


def _data_changed(data_id, data, hash_type, update):
    key = 'reactive.data_changed.%s' % data_id
    alg = getattr(hashlib, hash_type)
    serialized = json.dumps(data, sort_keys=True).encode('utf8')
    old_hash = unitdata.kv().get(key)
    new_hash = alg(serialized).hexdigest()
    if update:
        unitdata.kv().set(key, new_hash)
    return old_hash != new_hash


def data_changed(data_id, data, hash_type='md5'):
    """
    Check if the given set of data has changed since the previous call.

    This works by hashing the JSON-serialization of the data.  Note that,
    while the data will be serialized using ``sort_keys=True``, some types
    of data structures, such as sets, may lead to false positivies.

    :param str data_id: Unique identifier for this set of data.
    :param data: JSON-serializable data.
    :param str hash_type: Any hash algorithm supported by :mod:`hashlib`.
    """
    return _data_changed(data_id, data, hash_type, True)


def is_data_changed(data_id, data, hash_type='md5'):
    """
    Check if the given set of data has changed since the last time
    `data_changed` was called.

    That is, this is a non-destructive way to check if the data has changed.

    :param str data_id: Unique identifier for this set of data.
    :param data: JSON-serializable data.
    :param str hash_type: Any hash algorithm supported by :mod:`hashlib`.
    """
    return _data_changed(data_id, data, hash_type, False)


def _hook(hook_patterns):
    dispatch_phase = unitdata.kv().get('reactive.dispatch.phase')
    return dispatch_phase == 'hooks' and any_hook(*hook_patterns)


def _restricted_hook(hook_name):
    current_hook = hookenv.hook_name()
    dispatch_phase = unitdata.kv().get('reactive.dispatch.phase')
    return dispatch_phase == 'restricted' and current_hook == hook_name


def _when_all(flags):
    dispatch_phase = unitdata.kv().get('reactive.dispatch.phase')
    return dispatch_phase == 'other' and all_flags_set(*flags)


def _when_any(flags):
    dispatch_phase = unitdata.kv().get('reactive.dispatch.phase')
    return dispatch_phase == 'other' and any_flags_set(*flags)


def _when_none(flags):
    dispatch_phase = unitdata.kv().get('reactive.dispatch.phase')
    return dispatch_phase == 'other' and not any_flags_set(*flags)


def _when_not_all(flags):
    dispatch_phase = unitdata.kv().get('reactive.dispatch.phase')
    return dispatch_phase == 'other' and not all_flags_set(*flags)



# Copyright 2014-2017 Canonical Limited.
#
# This file is part of charms.reactive.
#
# charms.reactive is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import sys
import errno
import shutil
import tempfile
import unittest
import subprocess
from subprocess import Popen
from contextlib import contextmanager

import mock
from nose.plugins.attrib import attr

from charmhelpers.core import unitdata
from charms import reactive


class TestFlagWatch(unittest.TestCase):
    def setUp(self):
        self.key = reactive.bus.FlagWatch.key
        store_p = mock.patch.object(reactive.bus.FlagWatch, '_store')
        store = store_p.start()
        self.addCleanup(store_p.stop)
        self._data = {}
        store().get = self._data.get
        store().set = self._data.__setitem__
        store().unset = lambda k: self._data.pop(k, None)

    @property
    def data(self):
        return self._data.get(reactive.bus.FlagWatch.key, None)

    def test_reset(self):
        reactive.bus.FlagWatch.change('foo')
        reactive.bus.FlagWatch.reset()
        self.assertIsNone(self.data)

    def test_watch(self):
        reactive.bus.FlagWatch.reset()
        assert reactive.bus.FlagWatch.watch('foo', ['foos']), 'iter 0 (reset)'

        reactive.bus.FlagWatch.iteration(0)
        assert reactive.bus.FlagWatch.watch('foo', ['foos']), 'iter 0'

        reactive.bus.FlagWatch.iteration(1)
        assert not reactive.bus.FlagWatch.watch('foo', ['foos']), 'iter 1'

        reactive.bus.FlagWatch.change('foos')
        assert not reactive.bus.FlagWatch.watch('foo', ['foos']), 'uncommitted'

        reactive.bus.FlagWatch.commit()
        assert reactive.bus.FlagWatch.watch('foo', ['foos']), 'committed'

        reactive.bus.FlagWatch.change('bars')
        reactive.bus.FlagWatch.commit()
        assert not reactive.bus.FlagWatch.watch('foo', ['foos']), 'non-watched change'
        assert reactive.bus.FlagWatch.watch('bar', ['bars']), 'other watcher change'
        assert reactive.bus.FlagWatch.watch('foobar', ['foos', 'bars']), 'multi-watcher'

        reactive.bus.FlagWatch.commit()
        assert not reactive.bus.FlagWatch.watch('bar', ['bars']), 'already seen'
        assert not reactive.bus.FlagWatch.watch('foobar', ['foos', 'bars']), 'already seen multi'

        reactive.bus.FlagWatch.change('foos')
        reactive.bus.FlagWatch.change('bars')
        reactive.bus.FlagWatch.commit()
        assert reactive.bus.FlagWatch.watch('foo', ['foos']), 'multi-change, foo'
        assert reactive.bus.FlagWatch.watch('bar', ['bars']), 'multi-change, bar'
        assert reactive.bus.FlagWatch.watch('foobar', ['foos', 'bars']), 'multi-change, multi'
        assert not reactive.bus.FlagWatch.watch('qux', ['quxs']), 'multi-change, other'

        reactive.bus.FlagWatch.commit()
        assert not reactive.bus.FlagWatch.watch('foo', ['foos']), 'pre-reset'
        reactive.bus.FlagWatch.reset()
        assert reactive.bus.FlagWatch.watch('foo', ['foos']), 'post-reset'

    def test_change(self):
        reactive.bus.FlagWatch.change('foo')
        reactive.bus.FlagWatch.change('bar')
        self.assertEqual(self.data, {
            'iteration': 0,
            'pending': ['foo', 'bar'],
            'changes': [],
        })

    def test_commit(self):
        reactive.bus.FlagWatch.change('foo')
        reactive.bus.FlagWatch.change('bar')
        reactive.bus.FlagWatch.commit()
        self.assertEqual(self.data, {
            'iteration': 0,
            'pending': [],
            'changes': ['foo', 'bar'],
        })


class TestHandler(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_db_dir = tempfile.mkdtemp()
        cls.test_db = os.path.join(cls.test_db_dir, 'test-state.db')
        unitdata._KV = cls.kv = unitdata.Storage(cls.test_db)
        cls._log = mock.patch('charmhelpers.core.hookenv.log')
        cls.log = cls._log.start()
        cls._log_opts = mock.patch.dict(reactive.bus.LOG_OPTS, register=True)
        cls._log_opts.start()
        cls._charm_dir = mock.patch('charmhelpers.core.hookenv.charm_dir', lambda: 'charm_dir')
        cls._charm_dir.start()
        if not hasattr(cls, 'assertItemsEqual'):
            cls.assertItemsEqual = cls.assertCountEqual

    @classmethod
    def tearDownClass(cls):
        cls._log_opts.stop()
        cls._log.stop()
        cls._charm_dir.stop()
        cls.kv.close()
        unitdata._KV = None
        shutil.rmtree(cls.test_db_dir)

    def tearDown(self):
        reactive.bus.Handler.clear()
        self.kv.cursor.execute('delete from kv')

    def test_get(self):
        def test_action():
            pass

        handler = reactive.bus.Handler.get(test_action)
        self.assertIsInstance(handler, reactive.bus.Handler)
        handler2 = reactive.bus.Handler.get(test_action)
        self.assertIs(handler2, handler)

    def test_add_call_test(self):
        action = mock.Mock(name='action')
        pred1 = mock.Mock(name='pred1', return_value=True)
        pred2 = mock.Mock(name='pred2', return_value=False)

        def test_action(*args):
            action(*args)

        handler = reactive.bus.Handler.get(test_action)

        handler.add_predicate(pred1)
        assert handler.test()
        handler.add_predicate(pred2)
        assert not handler.test()
        pred2.return_value = True
        assert handler.test()
        assert not action.called

        handler.register_flags(['state1', 'state2'])
        assert handler.test()
        reactive.bus.FlagWatch.iteration(1)
        assert not handler.test()
        reactive.bus.FlagWatch.change('state1')
        assert not handler.test()
        reactive.bus.FlagWatch.commit()
        assert handler.test()

    def test_args(self):
        def test_action():
            pass

        handler = reactive.bus.Handler.get(test_action)
        handler.add_args(['arg1', 'arg2'])
        handler.add_args(arg.upper() for arg in ['arg3', 'arg4'])
        arg5 = mock.MagicMock()
        arg5.__iter__.return_value = ['arg5']
        handler.add_args(arg5)
        assert not arg5.__iter__.called
        self.assertEqual(handler._get_args(), ['arg1', 'arg2', 'ARG3', 'ARG4', 'arg5'])
        assert arg5.__iter__.called

    def test_invoke(self):
        action = mock.Mock(name='action')

        def test_action(*args):
            action(*args)

        handler = reactive.bus.Handler.get(test_action)
        handler.add_args(['arg1', 'arg2'])
        handler.invoke()
        action.assert_called_once_with('arg1', 'arg2')


class TestExternalHandler(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_db_dir = tempfile.mkdtemp()
        cls.test_db = os.path.join(cls.test_db_dir, 'test-state.db')
        unitdata._KV = cls.kv = unitdata.Storage(cls.test_db)
        cls._log = mock.patch('charmhelpers.core.hookenv.log')
        cls.log = cls._log.start()
        cls._log_opts = mock.patch.dict(reactive.bus.LOG_OPTS, register=True)
        cls._log_opts.start()
        cls._charm_dir = mock.patch('charmhelpers.core.hookenv.charm_dir', lambda: 'charm_dir')
        cls._charm_dir.start()
        if not hasattr(cls, 'assertItemsEqual'):
            cls.assertItemsEqual = cls.assertCountEqual

    @classmethod
    def tearDownClass(cls):
        cls._log_opts.stop()
        cls._log.stop()
        cls._charm_dir.stop()
        cls.kv.close()
        unitdata._KV = None
        shutil.rmtree(cls.test_db_dir)

    def tearDown(self):
        reactive.bus.Handler.clear()
        self.kv.cursor.execute('delete from kv')

    def test_register(self):
        handler1 = reactive.bus.ExternalHandler.register('filepath')
        handler2 = reactive.bus.ExternalHandler.register('filepath')
        handler3 = reactive.bus.ExternalHandler.register('filepath2')
        self.assertIs(handler1, handler2)
        self.assertIsNot(handler2, handler3)

    @mock.patch.object(os, 'environ', 'env')
    @mock.patch.object(reactive.bus.subprocess, 'Popen')
    def test_test(self, Popen):
        handler = reactive.bus.ExternalHandler('filepath')
        Popen.return_value.communicate.return_value = ('output', None)

        Popen.return_value.returncode = 0
        assert handler.test()

        Popen.return_value.returncode = 1
        assert not handler.test()
        Popen.assert_called_with(['filepath', '--test'], stdout=reactive.bus.subprocess.PIPE, env='env')

        e = Popen.side_effect = OSError()
        e.errno = errno.ENOEXEC
        self.assertRaises(reactive.bus.BrokenHandlerException, handler.test)
        e.errno = errno.ENOENT
        self.assertRaises(OSError, handler.test)

    @mock.patch.object(os, 'environ', 'env')
    @mock.patch.object(reactive.bus.subprocess, 'check_call')
    def test_invoke(self, check_call):
        handler = reactive.bus.ExternalHandler('filepath')
        handler._test_output = 'output'
        handler.invoke()
        check_call.assert_called_once_with(['filepath', '--invoke', 'output'], env='env')

        check_call.reset_mock()
        handler.invoke()
        check_call.assert_called_once_with(['filepath', '--invoke', 'output'], env='env')


class TestReactiveBus(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_db_dir = tempfile.mkdtemp()
        cls.test_db = os.path.join(cls.test_db_dir, 'test-state.db')
        unitdata._KV = cls.kv = unitdata.Storage(cls.test_db)
        cls._log = mock.patch('charmhelpers.core.hookenv.log')
        cls.log = cls._log.start()
        cls._charm_dir = mock.patch('charmhelpers.core.hookenv.charm_dir', lambda: 'charm_dir')
        cls._charm_dir.start()
        if not hasattr(cls, 'assertItemsEqual'):
            cls.assertItemsEqual = cls.assertCountEqual

    @classmethod
    def tearDownClass(cls):
        cls._log.stop()
        cls._charm_dir.stop()
        cls.kv.close()
        unitdata._KV = None
        shutil.rmtree(cls.test_db_dir)

    def tearDown(self):
        reactive.bus.Handler.clear()
        self.kv.cursor.execute('delete from kv')

    @mock.patch.object(reactive.bus.FlagWatch, 'change')
    def test_set_get_clear_flag(self, change):
        class states(reactive.flags.StateList):
            qux = reactive.flags.State('qux')

        reactive.set_flag('foo')
        reactive.set_flag('bar.ready', 'bar')
        reactive.set_flag(states.qux)
        self.assertEqual(reactive.flags.get_states(), {
            'foo': None,
            'bar.ready': 'bar',
            'qux': None,
        })
        reactive.clear_flag('foo')
        reactive.clear_flag('bar.ready')
        self.assertEqual(reactive.flags.get_states(), {
            'qux': None,
        })
        self.assertEqual(change.call_args_list, [
            mock.call('foo'),
            mock.call('bar.ready'),
            mock.call('qux'),
            mock.call('foo'),
            mock.call('bar.ready'),
        ])

    def test_dispatch(self):
        calls = []

        @reactive.when('foo')
        def foo():
            calls.append('foo')
            reactive.set_flag('bar')

        @reactive.when('bar')
        @reactive.only_once
        def bar():
            calls.append('bar')

        @reactive.when('foo', 'bar')
        def both():
            calls.append('both')
            reactive.set_flag('qux')

        @reactive.when('qux')
        def qux():
            calls.append('qux')

        @reactive.when('foo')
        def foo2():
            calls.append('foo2')

        @reactive.when('bar')
        def bar2():
            calls.append('bar2')

        reactive.bus.dispatch()
        self.assertEqual(calls, [])

        reactive.set_flag('foo')
        reactive.bus.dispatch()
        self.assertItemsEqual(calls[0:2], [
            'foo',
            'foo2',
        ])
        self.assertItemsEqual(calls[2:5], [
            'bar',
            'bar2',
            'both',
        ])
        self.assertItemsEqual(calls[5:], [
            'qux',
        ])

        # ensure that subsequent hooks will re-trigger all matching handlers (except only_once)
        calls = []
        reactive.bus.dispatch()
        self.assertItemsEqual(calls, [
            'foo',
            'both',
            'qux',
            'foo2',
            'bar2',
        ])

    @mock.patch.object(reactive.bus.Handler, 'get_handlers')
    def test_dispatch_remove(self, get_handlers):
        a = mock.Mock(name='a')
        a1 = lambda: a('h1') and reactive.clear_flag('foo')
        a2 = lambda: a('h2')
        a3 = lambda: a('h3')
        reactive.decorators.when('foo')(a1)
        reactive.decorators.when('foo')(a2)
        reactive.decorators.when('bar')(a3)
        h1 = reactive.bus.Handler.get(a1)
        h2 = reactive.bus.Handler.get(a2)
        h3 = reactive.bus.Handler.get(a3)

        reactive.set_flag('foo')
        reactive.set_flag('bar')
        get_handlers.return_value = [h1, h2, h3]
        reactive.bus.dispatch()
        self.assertEqual(a.call_args_list, [
            mock.call('h1'),
            mock.call('h3'),
        ])

        a.reset_mock()
        reactive.set_flag('foo')
        get_handlers.return_value = [h2, h1, h3]
        reactive.bus.dispatch()
        self.assertEqual(a.call_args_list, [
            mock.call('h2'),
            mock.call('h1'),
            mock.call('h3'),
        ])

    @mock.patch.object(reactive.bus.hookenv, 'hook_name')
    @mock.patch.object(reactive.bus.Handler, 'get_handlers')
    def test_dispatch_hook(self, get_handlers, hook_name):
        hook_name.return_value = 'fook'
        reactive.set_flag('foos')
        a = mock.Mock(name='a')
        a1 = lambda: a('h1')
        a2 = lambda: a('h2')
        reactive.decorators.when('foos')(a1)
        reactive.decorators.hook('fook')(a2)
        h1 = reactive.bus.Handler.get(a1)
        h2 = reactive.bus.Handler.get(a2)

        get_handlers.return_value = [h1, h2]
        reactive.bus.dispatch()
        self.assertEqual(a.call_args_list, [
            mock.call('h2'),
            mock.call('h1'),
        ])

        a.reset_mock()
        get_handlers.return_value = [h2, h1]
        reactive.bus.dispatch()
        self.assertEqual(a.call_args_list, [
            mock.call('h2'),
            mock.call('h1'),
        ])

    @mock.patch.dict('sys.modules')
    @mock.patch('charmhelpers.core.hookenv.charm_dir')
    def test_discover(self, charm_dir):
        test_dir = os.path.dirname(__file__)
        charm_dir.return_value = os.path.join(test_dir, 'data')

        self.assertEqual(len(reactive.bus.Handler.get_handlers()), 0)
        reactive.bus.discover()
        self.assertEqual(len(reactive.bus.Handler.get_handlers()), 19)

        # The path is extended so discovered modules can perform
        # absolute and relative imports as expected.
        self.assertListEqual(sys.path[-2:],
                             [charm_dir(), charm_dir() + '/hooks'])
        sys.path.pop()  # Repair sys.path
        sys.path.pop()

    @attr('slow')
    @mock.patch.dict('sys.modules')
    @mock.patch('charmhelpers.core.hookenv.relation_to_role_and_interface')
    @mock.patch('subprocess.check_call')
    @mock.patch('subprocess.Popen')
    @mock.patch('charmhelpers.core.hookenv.relation_type')
    @mock.patch('charmhelpers.core.hookenv.hook_name')
    @mock.patch('charmhelpers.core.hookenv.charm_dir')
    def test_full_stack(self, charm_dir, hook_name, relation_type, mPopen, mcheck_call, rtrai):
        test_dir = os.path.dirname(__file__)
        charm_dir.return_value = os.path.join(test_dir, 'data')
        hook_name.return_value = 'config-changed'
        relation_type.return_value = None

        mPopen.stdout = []
        mPopen.stderr = []

        def capture_output(proc):
            stdout, stderr = Popen.communicate(proc)
            mPopen.stdout.extend((stdout or b'').decode('utf-8').splitlines())
            mPopen.stderr.extend((stderr or b'').decode('utf-8').splitlines())
            return stdout, stderr

        def mock_Popen(*args, **kwargs):
            kwargs['stderr'] = subprocess.PIPE
            proc = Popen(*args, **kwargs)
            proc.communicate = lambda: capture_output(proc)
            return proc

        def mock_check_call(*args, **kwargs):
            proc = mock_Popen(*args, **kwargs)
            _, _ = proc.communicate()
            if proc.returncode != 0:
                raise subprocess.CalledProcessError(proc.returncode,
                                                    args[0])

        mPopen.side_effect = mock_Popen
        mcheck_call.side_effect = mock_check_call
        with mock.patch.dict(os.environ, {
            'PATH': os.pathsep.join([
                os.path.dirname(sys.executable),  # for /usr/bin/env python
                ':%s' % os.path.join(test_dir, '..', 'bin'),  # for chlp
                os.environ['PATH'],
            ]),
            'PYTHONPATH': os.pathsep.join(sys.path),
            'UNIT_STATE_DB': self.test_db,
            'CHARM_DIR': charm_dir.return_value,
        }):
            self.assertEqual(len(reactive.bus.Handler.get_handlers()), 0)
            reactive.bus.discover()
            self.assertEqual(len(reactive.bus.Handler.get_handlers()), 19)

            reactive.set_flag('test')
            reactive.set_flag('to-remove')
            assert not reactive.helpers.any_flags_set('top-level', 'nested', 'test-rel.ready', 'relation', 'bash')
            reactive.bus.dispatch()
            assert reactive.helpers.all_flags_set('top-level')
            assert reactive.helpers.all_flags_set('nested')
            assert not reactive.helpers.any_flags_set('relation')
            assert not reactive.helpers.any_flags_set('test-rel.ready')
            assert not reactive.helpers.any_flags_set('top-level-repeat')
            assert reactive.helpers.all_flags_set('bash-when')
            assert not reactive.helpers.all_flags_set('bash-when-repeat')
            assert not reactive.helpers.all_flags_set('bash-when-neg')
            assert reactive.helpers.all_flags_set('bash-when-any')
            assert not reactive.helpers.all_flags_set('bash-when-any-repeat')
            assert reactive.helpers.all_flags_set('bash-when-not')
            assert not reactive.helpers.all_flags_set('bash-when-not-repeat')
            assert not reactive.helpers.all_flags_set('bash-when-not-neg')
            assert reactive.helpers.all_flags_set('bash-when-not-all')
            assert not reactive.helpers.all_flags_set('bash-when-not-all-repeat')
            assert reactive.helpers.all_flags_set('bash-only-once')
            assert not reactive.helpers.all_flags_set('bash-only-once-repeat')
            assert not reactive.helpers.all_flags_set('bash-hook')
            assert not reactive.helpers.all_flags_set('bash-hook-repeat')
            assert reactive.helpers.all_flags_set('bash-multi')
            assert not reactive.helpers.all_flags_set('bash-multi-repeat')
            assert not reactive.helpers.all_flags_set('bash-multi-neg')
            assert not reactive.helpers.all_flags_set('bash-multi-neg2')
            invoked = ['test_when_not_all', 'test_when_any', 'test_when',
                       'test_when_not', 'test_multi', 'test_only_once']
            invoked.sort()
            stdout_invoked = mPopen.stdout[0].split(',')
            stdout_invoked.sort()
            self.assertEqual(stdout_invoked, invoked)
            stderr_invoked = next(
                line for line in mPopen.stderr
                if line.startswith("Will invoke:")
            )[13:].split(',')
            stderr_invoked.sort()
            self.assertEqual(stderr_invoked, invoked)
            for handler in invoked:
                assert 'Invoking bash reactive handler: %s' % handler in mPopen.stderr
            assert '++ charms.reactive set_flag bash-when-not-all' in mPopen.stderr
            bash_debug = False
            debug_debug = False
            for line in mPopen.stderr:
                if re.match(r'\+ REACTIVE_HANDLERS\[\$func]=.*/bash.sh:', line):
                    bash_debug = True
                if re.match(r'\+ REACTIVE_HANDLERS\[\$func]=.*/debug.sh:', line):
                    debug_debug = True
            assert not bash_debug, 'CHARMS_REACTIVE_TRACE not enabled but has debug output'
            assert debug_debug, 'CHARMS_REACTIVE_TRACE enabled but missing debug output'

            hook_name.return_value = 'test-rel-relation-joined'
            relation_type.return_value = 'test-rel'
            rtrai.return_value = ('requires', 'test')
            self.assertIsNotNone(reactive.relations.relation_factory('test-rel'))
            self.assertIsNotNone(reactive.relations.relation_factory('test-rel').from_name('test'))
            with mock.patch.dict(os.environ, {'JUJU_HOOK_NAME': 'test-rel-relation-joined'}):
                reactive.bus.dispatch()
            assert reactive.helpers.all_flags_set('test-rel.ready')
            assert reactive.helpers.all_flags_set('relation')
            assert reactive.helpers.all_flags_set('test-remove-not')
            assert reactive.helpers.all_flags_set('bash-hook')
            assert not reactive.helpers.all_flags_set('bash-hook-repeat')
            print('\n'.join(mPopen.stderr))
            assert reactive.helpers.all_flags_set('bash-when-repeat')
            assert not reactive.helpers.all_flags_set('bash-only-once-repeat')

        # The path is extended so discovered modules can perform
        # absolute and relative imports as expected.
        self.assertListEqual(sys.path[-2:],
                             [charm_dir(), charm_dir() + '/hooks'])
        sys.path.pop()  # Repair sys.path
        sys.path.pop()

    @attr('slow')
    @mock.patch('charmhelpers.core.hookenv.log')
    def test_full_stack_with_tracing(self, log):
        self.addCleanup(reactive.trace.install_tracer, reactive.trace.NullTracer())
        tracer = mock.Mock(wraps=reactive.trace.LogTracer())
        reactive.trace.install_tracer(tracer)
        self.test_full_stack()
        log.assert_called()
        tracer.start_dispatch.assert_called()
        tracer.start_dispatch_phase.assert_any_call('hooks', mock.ANY)
        tracer.start_dispatch_phase.assert_any_call('other', mock.ANY)
        tracer.start_dispatch_iteration.assert_any_call(0, mock.ANY)

    @mock.patch.object(reactive.bus.importlib, 'import_module')
    def test_load_module_py3(self, import_module):
        import_module.side_effect = lambda x: x
        self.assertEqual(reactive.bus._load_module('/x/reactive',
                                                   '/x/reactive/file1.py'),
                         'reactive.file1')
        import_module.assert_called_once_with('reactive.file1')

        import_module.reset_mock()
        mod = reactive.bus._load_module('/x/hooks/relations',
                                        '/x/hooks/relations/iface/role.py')
        self.assertEqual(mod, 'relations.iface.role')
        import_module.assert_called_once_with('relations.iface.role')

        import_module.reset_mock()
        mod = reactive.bus._load_module('/x/hooks/reactive',
                                        '/x/hooks/reactive/sub/__init__.py')
        self.assertEqual(mod, 'reactive.sub')
        import_module.assert_called_once_with('reactive.sub')

    def test_load_module_really(self):
        # Look Ma, no mocks!
        here = os.path.dirname(__file__)
        root_path = os.path.join(here, 'data')

        # Path manipulation normally done by bus.discover()
        ext_path = [root_path, os.path.join(root_path, 'hooks')]
        with extended_sys_path(ext_path):
            top_mod = reactive.bus._load_module(os.path.join(root_path,
                                                             'reactive'),
                                                os.path.join(root_path,
                                                             'reactive',
                                                             'top_level.py'))
            sub_mod = reactive.bus._load_module(os.path.join(root_path,
                                                             'reactive'),
                                                os.path.join(root_path,
                                                             'reactive',
                                                             'nested',
                                                             'nested.py'))
            hyp_mod = reactive.bus._load_module(os.path.join(root_path,
                                                             'hooks',
                                                             'relations'),
                                                os.path.join(root_path,
                                                             'hooks',
                                                             'relations',
                                                             'hyphen-ated',
                                                             'peer.py'))

        self.assertEqual(top_mod.test_marker, 'top level')
        self.assertEqual(sub_mod.test_marker, 'nested')
        self.assertEqual(hyp_mod.test_marker, 'hyphenated-relation')

        self.assertIn('reactive.top_level', sys.modules)
        self.assertIn('reactive.nested.nested', sys.modules)
        self.assertIs(top_mod, sys.modules['reactive.top_level'])
        self.assertIs(sub_mod, sys.modules['reactive.nested.nested'])

    @mock.patch.object(reactive.bus.ExternalHandler, 'register')
    @mock.patch.object(reactive.bus, '_is_external_handler')
    @mock.patch.object(reactive.bus, '_load_module')
    def test_register_handlers_from_file(self, _load_module, _is_external_handler, register):
        reactive.bus._register_handlers_from_file('reactive',
                                                  'reactive/foo.py')
        _load_module.assert_called_once_with('reactive',
                                             'reactive/foo.py')
        _is_external_handler.return_value = True
        reactive.bus._register_handlers_from_file('reactive', 'reactive/foo')
        _is_external_handler.assert_called_once_with('reactive/foo')
        register.assert_called_once_with('reactive/foo')

        register.reset_mock()
        _is_external_handler.return_value = False
        reactive.bus._register_handlers_from_file('hooks/relations',
                                                  'hooks/relations/foo/README.md')
        assert not register.called


@contextmanager
def extended_sys_path(extras_list):
    extras_list = [p for p in extras_list if p not in sys.path]
    sys.path.extend(extras_list)
    try:
        yield
    finally:
        for p in extras_list:
            sys.path.remove(p)


if __name__ == '__main__':
    unittest.main()



from charms.reactive import Endpoint, when


class TestAltRequires(Endpoint):
    invocations = []

    @when('endpoint.{endpoint_name}.joined')
    def handle_joined(self):
        self.invocations.append('joined: {}'.format(self.endpoint_name))

    @when('endpoint.{endpoint_name}.changed')
    def handle_changed(self):
        self.invocations.append('changed: {}'.format(self.endpoint_name))

    @when('endpoint.{endpoint_name}.changed.foo')
    def handle_changed_foo(self):
        self.invocations.append('changed.foo: {}'.format(self.endpoint_name))



from charmhelpers.cli import cmdline
from charmhelpers.core import hookenv
from charmhelpers.core import unitdata

from charms.reactive.bus import FlagWatch
from charms.reactive.trace import tracer


__all__ = [
    'set_flag',
    'clear_flag',
    'toggle_flag',
    'register_trigger',
    'is_flag_set',
    'all_flags_set',
    'any_flags_set',
    'get_flags',
    'get_unset_flags',
    'set_state',  # DEPRECATED
    'remove_state',  # DEPRECATED
    'toggle_state',  # DEPRECATED
    'is_state',  # DEPRECATED
    'all_states',  # DEPRECATED
    'get_states',  # DEPRECATED
    'any_states',  # DEPRECATED
]

TRIGGER_CALLBACKS = {}


class State(str):
    """
    .. deprecated:: 0.5.0
       Use flag instead

    A reactive state that can be set.

    States are essentially just strings, but this class should be used to enable them
    to be discovered and introspected, for documentation, composition, or linting.

    This should be used with :class:`StateList`.
    """
    pass


class StateList(object):
    """
    .. deprecated:: 0.5.0
       use flag instead

    Base class for a set of states that can be set by a relation or layer.

    This class should be used so that they can be discovered and introspected,
    for documentation, composition, or linting.

    Example usage::

        class MyRelation(RelationBase):
            class states(StateList):
                connected = State('{relation_name}.connected')
                available = State('{relation_name}.available')
    """
    pass


@cmdline.subcommand()
@cmdline.no_output
def set_flag(flag, value=None):
    """set_flag(flag)
    Set the given flag as active.

    :param str flag: Name of flag to set.

    .. note:: **Changes to flags are reset when a handler crashes.** Changes to
       flags happen immediately, but they are only persisted at the end of a
       complete and successful run of the reactive framework. All unpersisted
       changes are discarded when a hook crashes.
    """
    old_flags = get_flags()
    unitdata.kv().update({flag: value}, prefix='reactive.states.')
    if flag not in old_flags:
        tracer().set_flag(flag)
        FlagWatch.change(flag)
        trigger = _get_trigger(flag, None)
        for flag_name in trigger['set_flag']:
            set_flag(flag_name)
        for flag_name in trigger['clear_flag']:
            clear_flag(flag_name)
        for callback in trigger['callbacks']:
            callback()


@cmdline.subcommand()
@cmdline.no_output
def clear_flag(flag):
    """
    Clear / deactivate a flag.

    :param str flag: Name of flag to set.

    .. note:: **Changes to flags are reset when a handler crashes.** Changes to
       flags happen immediately, but they are only persisted at the end of a
       complete and successful run of the reactive framework. All unpersisted
       changes are discarded when a hook crashes.
    """
    old_flags = get_flags()
    unitdata.kv().unset('reactive.states.%s' % flag)
    unitdata.kv().set('reactive.dispatch.removed_state', True)
    if flag in old_flags:
        tracer().clear_flag(flag)
        FlagWatch.change(flag)
        trigger = _get_trigger(None, flag)
        for flag_name in trigger['set_flag']:
            set_flag(flag_name)
        for flag_name in trigger['clear_flag']:
            clear_flag(flag_name)
        for callback in trigger['callbacks']:
            callback()


@cmdline.subcommand()
@cmdline.no_output
def toggle_flag(flag, should_set):
    """
    Helper that calls either :func:`set_flag` or :func:`clear_flag`,
    depending on the value of `should_set`.

    Equivalent to::

        if should_set:
            set_flag(flag)
        else:
            clear_flag(flag)

    :param str flag: Name of flag to toggle.
    :param bool should_set: Whether to set the flag, or clear it.

    .. note:: **Changes to flags are reset when a handler crashes.** Changes to
       flags happen immediately, but they are only persisted at the end of a
       complete and successful run of the reactive framework. All unpersisted
       changes are discarded when a hook crashes.
    """
    if should_set:
        set_flag(flag)
    else:
        clear_flag(flag)


@cmdline.subcommand()
@cmdline.no_output
def register_trigger(when=None, when_not=None, set_flag=None, clear_flag=None, callback=None):
    """
    Register a trigger to set or clear a flag when a given flag is set.

    Note: Flag triggers are handled at the same time that the given flag is set.

    :param str when: Flag to trigger on when it is set.
    :param str when_not: Flag to trigger on when it is cleared.
    :param str set_flag: If given, this flag will be set when the relevant flag is changed.
    :param str clear_flag: If given, this flag will be cleared when the relevant flag is changed.
    :param str callback: If given, this callback will be invoked when the relevant flag is changed.

    Note: Exactly one of either `when` or `when_not`, and at least one of
    `set_flag`, `clear_flag`, or `callback` must be provided.
    """
    if not any((when, when_not)):
        raise ValueError('Must provide one of when or when_not')
    if all((when, when_not)):
        raise ValueError('Only one of when or when_not can be provided')
    if not any((set_flag, clear_flag, callback)):
        raise ValueError('Must provide at least one of set_flag, clear_flag, or callback')
    trigger = _get_trigger(when, when_not)
    if set_flag and set_flag not in trigger['set_flag']:
        trigger['set_flag'].append(set_flag)
    if clear_flag and clear_flag not in trigger['clear_flag']:
        trigger['clear_flag'].append(clear_flag)
    if callback and callback not in trigger['callbacks']:
        trigger['callbacks'].append(callback)
    _save_trigger(when, when_not, trigger)


def _get_trigger(when, when_not):
    if when is not None:
        key = 'reactive.flag_set_triggers.{}'.format(when)
    elif when_not is not None:
        key = 'reactive.flag_clear_triggers.{}'.format(when_not)
    triggers = unitdata.kv().get(key, {
        'set_flag': [],
        'clear_flag': [],
    })
    triggers['callbacks'] = TRIGGER_CALLBACKS.get(key, [])
    return triggers


def _save_trigger(when, when_not, data):
    if when is not None:
        key = 'reactive.flag_set_triggers.{}'.format(when)
    elif when_not is not None:
        key = 'reactive.flag_clear_triggers.{}'.format(when_not)
    TRIGGER_CALLBACKS[key] = data.pop('callbacks')
    return unitdata.kv().set(key, data)


@cmdline.subcommand()
@cmdline.test_command
def is_flag_set(flag):
    """Assert that a flag is set"""
    return any_flags_set(flag)


@cmdline.subcommand()
@cmdline.test_command
def all_flags_set(*desired_flags):
    """Assert that all desired_flags are set"""
    active_flags = get_flags()
    return all(flag in active_flags for flag in desired_flags)


@cmdline.subcommand()
@cmdline.test_command
def any_flags_set(*desired_flags):
    """Assert that any of the desired_flags are set"""
    active_flags = get_flags()
    return any(flag in active_flags for flag in desired_flags)


@cmdline.subcommand()
def get_flags():
    """
    Return a list of all flags which are set.
    """
    flags = unitdata.kv().getrange('reactive.states.', strip=True) or {}
    return sorted(flags.keys())


@cmdline.subcommand()
def get_unset_flags(*desired_flags):
    """Check if any of the provided flags missing and return them if so.

    :param desired_flags: list of reactive flags
    :type desired_flags: non-keyword args, str
    :returns: list of unset flags filtered from the parameters shared
    :rtype: List[str]
    """
    flags = unitdata.kv().getrange('reactive.states.', strip=True) or {}
    return sorted(set(desired_flags) - flags.keys())


def _get_flag_value(flag, default=None):
    return unitdata.kv().get('reactive.states.%s' % flag, default)


# DEPRECATED

@cmdline.subcommand()
@cmdline.no_output
def set_state(state, value=None):
    """
    .. deprecated:: 0.5.0
       Alias of :func:`set_flag`.
    """
    set_flag(state, value)


@cmdline.subcommand()
@cmdline.no_output
def remove_state(state):
    """
    .. deprecated:: 0.5.0
       Alias of :func:`clear_flag`"""
    clear_flag(state)


def toggle_state(state, should_set):
    """
    .. deprecated:: 0.5.0
       Alias of :func:`toggle_flag`"""
    toggle_flag(state, should_set)


@cmdline.subcommand()
@cmdline.test_command
def is_state(state):
    """
    .. deprecated:: 0.5.0
       Alias for :func:`is_flag_set`"""
    return is_flag_set(state)


@cmdline.subcommand()
@cmdline.test_command
def all_states(*desired_states):
    """
    .. deprecated:: 0.5.0
       Alias for :func:`all_flags_set`"""
    return all_flags_set(*desired_states)


@cmdline.subcommand()
@cmdline.test_command
def any_states(*desired_states):
    """
    .. deprecated:: 0.5.0
       Alias for :func:`any_flags_set`"""
    return any_flags_set(*desired_states)


@cmdline.subcommand()
def get_states():
    """
    .. deprecated:: 0.5.0
       Use :func:`get_flags` instead.

    Return a mapping of all active states to their values.
    """
    return unitdata.kv().getrange('reactive.states.', strip=True) or {}


def get_state(flag, default=None):
    """
    .. deprecated:: 0.5.0
       For internal use only.
    """
    return _get_flag_value(flag, default)


# INTERNAL

@hookenv.atstart
def _manage_automatic_flags():
    _manage_upgrade_flags()


def _manage_upgrade_flags():
    hook_name = hookenv.hook_name()

    if hook_name == 'pre-series-upgrade':
        set_flag('upgrade.series.in-progress')

    if hook_name == 'post-series-upgrade':
        clear_flag('upgrade.series.in-progress')


@hookenv.atexit
def _clear_triggers():
    unitdata.kv().unsetrange(prefix='reactive.flag_triggers.')  # old key
    unitdata.kv().unsetrange(prefix='reactive.flag_set_triggers.')
    unitdata.kv().unsetrange(prefix='reactive.flag_clear_triggers.')
